import pathlib
import inspect
import copy
import enum
import typeguard
import typing
from typing import Any, Literal

from glassesTools import annotation, aruco, gaze_worldref, json, marker as gt_marker, utils as gt_utils
from glassesTools.validation import DataQualityType, get_DataQualityType_explanation

from . import marker, plane, session, typed_dict_defaults, type_utils, utils


class AutoCodeSyncPoints(typed_dict_defaults.TypedDictDefault, total=False):
    markers         : set[gt_marker.MarkerID]
    max_gap_duration: int       = 4
    min_duration    : int       = 6

class AutoCodeEpisodes(typed_dict_defaults.TypedDictDefault, total=False):
    start_markers               : list[gt_marker.MarkerID]
    end_markers                 : list[gt_marker.MarkerID]
    max_gap_duration            : int       = 4
    max_intermarker_gap_duration: int       = 15
    min_duration                : int       = 6

class I2MCSettings(typed_dict_defaults.TypedDictDefault, total=False):
    # None where fields are set by code dynamically based on the data. When value applied here, it overrides this dynamic parameter setting
    freq            : float|None    = None
    windowtimeInterp: float         = .25       # s
    edgeSampInterp  : int           = 2
    maxdisp         : float         = 50        # mm
    windowtime      : float         = .2        # s
    steptime        : float         = .02       # s
    downsamples     : set[int]|None = None
    downsampFilter  : bool|None     = None
    chebyOrder      : int|None      = None
    maxerrors       : int           = 100
    cutoffstd       : float|None    = None
    onoffsetThresh  : float         = 3.
    maxMergeDist    : float         = 20        # mm
    maxMergeTime    : float         = 81        # ms
    minFixDur       : float         = 50        # ms

class CamMovementForEtSyncFunction(typed_dict_defaults.TypedDictDefault, total=False):
    module_or_file  : str
    function        : str
    parameters      : dict[str,Any]|None = None

class RgbColor(typing.NamedTuple):
    r: int = 0
    g: int = 0
    b: int = 0
json.register_type(json.TypeEntry(RgbColor, '__config.RgbColor__', lambda x: x._asdict(), lambda x: RgbColor(**x)))

class Study:
    default_json_file_name = 'study_def.json'

    @typeguard.typechecked(collection_check_strategy=typeguard.CollectionCheckStrategy.ALL_ITEMS)
    def __init__(self,
                 session_def                                    : session.SessionDefinition,
                 planes                                         : list[plane.Definition],
                 planes_per_episode                             : dict[annotation.Event,set[str]],
                 episodes_to_code                               : set[annotation.Event],
                 individual_markers                             : list[marker.Marker],
                 working_directory                              : str|pathlib.Path,

                 # setup with defaults
                 import_do_copy_video                           : bool                              = True,
                 import_source_dir_as_relative_path             : bool                              = False,
                 import_known_custom_eye_trackers               : list[str]|None                    = None,

                 overlay_video_gaze_vid_pos_color               : RgbColor                          = RgbColor(  0,255,  0),
                 overlay_video_gaze_world_pos_color             : RgbColor|None                     = RgbColor(255,  0,255),
                 overlay_video_gaze_vid_pos_radius              : int                               = 8,
                 overlay_video_gaze_world_pos_radius            : int                               = 5,
                 overlay_video_gaze_vid_pos_thickness           : int                               = 2,
                 overlay_video_gaze_world_pos_thickness         : int                               = -1,

                 sync_ref_recording                             : str|None                          = None,
                 sync_ref_do_time_stretch                       : bool|None                         = None,
                 sync_ref_stretch_which                         : Literal['ref','other']|None       = None,
                 sync_ref_average_recordings                    : set[str]|None                     = None,

                 get_cam_movement_for_et_sync_method            : Literal['','plane','function']    = '',
                 get_cam_movement_for_et_sync_function          : CamMovementForEtSyncFunction|None = None,
                 sync_et_to_cam_use_average                     : bool                              = True,

                 auto_code_sync_points                          : AutoCodeSyncPoints|None           = None,
                 auto_code_episodes                             : dict[Literal[
                                                                    annotation.Event.Sync_ET_Data,
                                                                    annotation.Event.Trial,
                                                                    annotation.Event.Validate],
                                                                  AutoCodeEpisodes]|None            = None,

                 export_output3D                                : bool                              = False,
                 export_output2D                                : bool                              = True,
                 export_only_code_marker_presence               : bool                              = True,

                 validate_do_global_shift                       : bool                              = True,
                 validate_max_dist_fac                          : float                             = .5,
                 validate_dq_types                              : set[DataQualityType]|None         = None,
                 validate_allow_dq_fallback                     : bool                              = False,
                 validate_include_data_loss                     : bool                              = False,
                 validate_I2MC_settings                         : I2MCSettings|None                 = None,
                 validate_dynamic_skip_first_duration           : float                             = .2,
                 validate_dynamic_max_gap_duration              : int                               = 4,
                 validate_dynamic_min_duration                  : int                               = 6,

                 mapped_video_make_which                               : set[str]|None                     = None,
                 mapped_video_recording_colors                         : dict[str,RgbColor]|None           = None,
                 mapped_video_projected_vidPos_color                   : RgbColor|None                     = RgbColor(255,255,  0),
                 mapped_video_projected_world_pos_color                : RgbColor|None                     = RgbColor(255,  0,255),
                 mapped_video_projected_left_ray_color                 : RgbColor|None                     = RgbColor(  0,  0,255),
                 mapped_video_projected_right_ray_color                : RgbColor|None                     = RgbColor(255,  0,  0),
                 mapped_video_projected_average_ray_color              : RgbColor|None                     = RgbColor(255,  0,255),
                 mapped_video_process_planes_for_all_frames            : bool                              = False,
                 mapped_video_process_annotations_for_all_recordings   : bool                              = True,
                 mapped_video_plane_marker_color                       : RgbColor|None                     = RgbColor(  0,255,  0),
                 mapped_video_recovered_plane_marker_color             : RgbColor|None                     = RgbColor(  0,255,255),
                 mapped_video_plane_axis_arm_length                    : float                             = 25,
                 mapped_video_individual_marker_color                  : RgbColor|None                     = RgbColor(255,  0,255),
                 mapped_video_individual_marker_axis_arm_length        : float                             = 25,
                 mapped_video_unexpected_marker_color                  : RgbColor|None                     = RgbColor(255,255,128),
                 mapped_video_rejected_marker_color                    : RgbColor|None                     = None,
                 mapped_video_show_sync_func_output                    : bool                              = True,
                 mapped_video_show_gaze_on_plane_in_which              : set[str]|None                     = None,
                 mapped_video_show_gaze_vec_in_which                   : set[str]|None                     = None,
                 mapped_video_show_camera_in_which                     : set[str]|None                     = None,
                 mapped_video_which_gaze_type_on_plane                 : gaze_worldref.Type                = gaze_worldref.Type.Scene_Video_Position,
                 mapped_video_which_gaze_type_on_plane_allow_fallback  : bool                              = True,
                 mapped_video_gaze_to_plane_margin                     : float                             = 0.25,

                 gui_num_workers                                : int                               = 2,

                 # not a class member
                 strict_check                                   : bool                              = True
                 ):
        self.session_def                                    = session_def
        self.planes                                         = planes
        self.planes_per_episode                             = planes_per_episode
        self.episodes_to_code                               = episodes_to_code
        self.individual_markers                             = individual_markers
        self.working_directory                              = working_directory

        self.import_do_copy_video                           = import_do_copy_video
        self.import_source_dir_as_relative_path             = import_source_dir_as_relative_path
        self.import_known_custom_eye_trackers               = import_known_custom_eye_trackers

        self.overlay_video_gaze_vid_pos_color               = overlay_video_gaze_vid_pos_color
        self.overlay_video_gaze_world_pos_color             = overlay_video_gaze_world_pos_color
        self.overlay_video_gaze_vid_pos_radius              = overlay_video_gaze_vid_pos_radius
        self.overlay_video_gaze_world_pos_radius            = overlay_video_gaze_world_pos_radius
        self.overlay_video_gaze_vid_pos_thickness           = overlay_video_gaze_vid_pos_thickness
        self.overlay_video_gaze_world_pos_thickness         = overlay_video_gaze_world_pos_thickness

        self.get_cam_movement_for_et_sync_method            = get_cam_movement_for_et_sync_method
        self.get_cam_movement_for_et_sync_function          = get_cam_movement_for_et_sync_function
        self.sync_et_to_cam_use_average                     = sync_et_to_cam_use_average

        self.sync_ref_recording                             = sync_ref_recording
        self.sync_ref_do_time_stretch                       = sync_ref_do_time_stretch
        self.sync_ref_stretch_which                         = sync_ref_stretch_which
        self.sync_ref_average_recordings                    = sync_ref_average_recordings

        self.auto_code_sync_points                          = auto_code_sync_points
        self.auto_code_episodes                             = auto_code_episodes

        self.export_output3D                                = export_output3D
        self.export_output2D                                = export_output2D
        self.export_only_code_marker_presence               = export_only_code_marker_presence

        self.validate_do_global_shift                       = validate_do_global_shift
        self.validate_max_dist_fac                          = validate_max_dist_fac
        self.validate_dq_types                              = validate_dq_types
        self.validate_allow_dq_fallback                     = validate_allow_dq_fallback
        self.validate_include_data_loss                     = validate_include_data_loss
        self.validate_I2MC_settings                         = validate_I2MC_settings
        self.validate_dynamic_skip_first_duration           = validate_dynamic_skip_first_duration
        self.validate_dynamic_max_gap_duration              = validate_dynamic_max_gap_duration
        self.validate_dynamic_min_duration                  = validate_dynamic_min_duration

        self.mapped_video_make_which                               = mapped_video_make_which
        self.mapped_video_recording_colors                         = mapped_video_recording_colors
        self.mapped_video_projected_vidPos_color                   = mapped_video_projected_vidPos_color
        self.mapped_video_projected_world_pos_color                = mapped_video_projected_world_pos_color
        self.mapped_video_projected_left_ray_color                 = mapped_video_projected_left_ray_color
        self.mapped_video_projected_right_ray_color                = mapped_video_projected_right_ray_color
        self.mapped_video_projected_average_ray_color              = mapped_video_projected_average_ray_color
        self.mapped_video_process_planes_for_all_frames            = mapped_video_process_planes_for_all_frames             # if True, all planes are processed for all frames, if False, only according to the planes_per_episode setup and the coding
        self.mapped_video_process_annotations_for_all_recordings   = mapped_video_process_annotations_for_all_recordings    # if True, all coded episodes for all planes of all recordings are processed (so e.g. if validation coded for one recording in the session, that plane is processed for all)
        self.mapped_video_plane_marker_color                       = mapped_video_plane_marker_color
        self.mapped_video_recovered_plane_marker_color             = mapped_video_recovered_plane_marker_color
        self.mapped_video_plane_axis_arm_length                    = mapped_video_plane_axis_arm_length
        self.mapped_video_individual_marker_color                  = mapped_video_individual_marker_color
        self.mapped_video_individual_marker_axis_arm_length        = mapped_video_individual_marker_axis_arm_length
        self.mapped_video_unexpected_marker_color                  = mapped_video_unexpected_marker_color
        self.mapped_video_rejected_marker_color                    = mapped_video_rejected_marker_color
        self.mapped_video_show_sync_func_output                    = mapped_video_show_sync_func_output
        self.mapped_video_show_gaze_on_plane_in_which              = mapped_video_show_gaze_on_plane_in_which
        self.mapped_video_show_gaze_vec_in_which                   = mapped_video_show_gaze_vec_in_which
        self.mapped_video_show_camera_in_which                     = mapped_video_show_camera_in_which
        self.mapped_video_which_gaze_type_on_plane                 = mapped_video_which_gaze_type_on_plane
        self.mapped_video_which_gaze_type_on_plane_allow_fallback  = mapped_video_which_gaze_type_on_plane_allow_fallback
        self.mapped_video_gaze_to_plane_margin                     = mapped_video_gaze_to_plane_margin                      # fraction of plane size, added to each side of the plane

        self.gui_num_workers                                = gui_num_workers

        self.check_valid(strict_check=strict_check)

    def check_valid(self, strict_check=True):
        if strict_check:
            self._check_session_def(strict_check)
            self._check_planes_per_episode(strict_check)
            self._check_episodes_to_code(strict_check)
            self._check_individual_markers(strict_check)
            self._check_sync_ref(strict_check)
            self._check_et_sync_method(strict_check)
            self._check_auto_coding_setup(strict_check)
            self._check_make_video(strict_check)

        # ensure typed dicts with defaults members are of the right class, and apply defaults
        to_check = {k:t for k in study_parameter_types if typed_dict_defaults.is_typeddictdefault(t:=gt_utils.unpack_none_union(study_parameter_types[k])[0])}
        for k in to_check:
            if getattr(self,k) is not None:
                setattr(self,k, to_check[k](getattr(self,k)))
                getattr(self,k).apply_defaults()
        if self.auto_code_episodes is not None:
            self.auto_code_episodes = {e: AutoCodeEpisodes(self.auto_code_episodes[e]) for e in self.auto_code_episodes}
            for e in self.auto_code_episodes:
                self.auto_code_episodes[e].apply_defaults()

    def _check_recordings(self, which: list[str]|None, field: str, strict_check) -> type_utils.ProblemDict:
        problems: type_utils.ProblemDict = {}
        if which is None:
            return problems
        missing_recs: list[str] = []
        for w in which:
            if not self._check_recording(w):
                if strict_check:
                    raise ValueError(f'Recording "{w}" not known, check {field} in the study configuration')
                else:
                    missing_recs.append(w)
        if missing_recs:
            problems[field] = f'Recording(s) {missing_recs[0] if len(missing_recs)==1 else missing_recs} not known'
            if isinstance(getattr(self,field),dict):
                type_utils.merge_problem_dicts(problems,{field: {r:f'Recording {r} not known' for r in missing_recs}})
        return problems

    def _check_recording(self, rec: str) -> bool:
        return any([r.name==rec for r in self.session_def.recordings])

    def _check_session_def(self, strict_check) -> type_utils.ProblemDict:
        # require at least one eye tracker recording
        if not any(r.type==session.RecordingType.Eye_Tracker for r in self.session_def.recordings):
            if strict_check:
                raise ValueError('At least one recording should be an eye tracker recording')
            else:
                return {'session_def': 'At least one recording should be an eye tracker recording'}
        return {}

    def _check_planes_per_episode(self, strict_check) -> type_utils.ProblemDict:
        problems: type_utils.ProblemDict = {}
        if not self.planes_per_episode:
            type_utils.merge_problem_dicts(problems, {'planes_per_episode': 'At minimum one episode should have a corresponding plane defined'})
        for e in self.planes_per_episode:
            missing_planes: list[str] = []
            for p in self.planes_per_episode[e]:
                if not any([p==pl.name for pl in self.planes]):
                    if strict_check:
                        raise ValueError(f'Plane {p} not known')
                    else:
                        missing_planes.append(p)
            if missing_planes:
                mp = '", "'.join(missing_planes)
                type_utils.merge_problem_dicts(problems, {'planes_per_episode': {e: f'Plane(s) "{mp}" not known.'}})

            # check correct number of planes is defined for the episode
            match e:
                case annotation.Event.Sync_Camera:
                    allow_one_plane = allow_more_than_one = False
                case annotation.Event.Sync_ET_Data:
                    if self.get_cam_movement_for_et_sync_method=='plane':
                        allow_one_plane = True
                        allow_more_than_one = False
                    else:
                        allow_one_plane = allow_more_than_one = False
                case annotation.Event.Validate:
                    allow_one_plane = True
                    allow_more_than_one = False
                case annotation.Event.Trial:
                    allow_one_plane = allow_more_than_one = True
            if not allow_one_plane:
                msg = f'No planes should be defined for a {annotation.tooltip_map[e]}. Remove entry, even if its empty.'
                if e==annotation.Event.Sync_ET_Data:
                    msg += ' Alternatively, you may want to set the get_cam_movement_for_et_sync_method on the main options panel to "Plane".'
                if strict_check:
                    raise ValueError(msg)
                else:
                    type_utils.merge_problem_dicts(problems, {'planes_per_episode': {e: msg}})
            elif not self.planes_per_episode[e]:
                msg = ('At least one' if allow_more_than_one else 'One')+f' plane should be defined for a {annotation.tooltip_map[e]}'
                if strict_check:
                    raise ValueError(msg)
                else:
                    type_utils.merge_problem_dicts(problems, {'planes_per_episode': {e: msg}})
            if not allow_more_than_one and len(self.planes_per_episode[e])>1:
                msg = f'Only one plane should be defined for a {annotation.tooltip_map[e]}'
                if strict_check:
                    raise ValueError(msg)
                else:
                    type_utils.merge_problem_dicts(problems, {'planes_per_episode': {e: msg}})

        for e in self.episodes_to_code:
            if e not in self.planes_per_episode and e!=annotation.Event.Sync_Camera and (e==annotation.Event.Sync_ET_Data and self.get_cam_movement_for_et_sync_method=='plane'):
                msg = f'{annotation.tooltip_map[e]}s are set up to be coded and require an associated plane, but no plane(s) are defined in planes_per_episode for {annotation.tooltip_map[e]}s'
                if strict_check:
                    raise ValueError(msg)
                else:
                    type_utils.merge_problem_dicts(problems, {'planes_per_episode': msg})
        return problems

    def _check_episodes_to_code(self, strict_check) -> type_utils.ProblemDict:
        problems: type_utils.ProblemDict = {}
        if not self.episodes_to_code:
            type_utils.merge_problem_dicts(problems, {'episodes_to_code': 'At minimum one episode should be selected to be coded'})
        if annotation.Event.Sync_ET_Data in self.episodes_to_code and self.get_cam_movement_for_et_sync_method=='':
            type_utils.merge_problem_dicts(problems, {'episodes_to_code': f'{annotation.tooltip_map[annotation.Event.Sync_ET_Data]} should not be listed in the episodes to be coded if there is no method for plane synchronization (get_cam_movement_for_et_sync_method) specified.'})

        for e in self.planes_per_episode:
            if e not in self.episodes_to_code:
                if strict_check:
                    raise ValueError(f'Plane(s) are defined in planes_per_episode for {annotation.tooltip_map[e]}s, but {annotation.tooltip_map[e]}s are not set up to be coded in episodes_to_code. Fix episodes_to_code.')
                else:
                    type_utils.merge_problem_dicts(problems, {'episodes_to_code': f'Plane(s) are defined in planes_per_episode for {annotation.tooltip_map[e]}s, but {annotation.tooltip_map[e]}s are not set up to be coded'})
                    type_utils.merge_problem_dicts(problems, {'planes_per_episode': {e: f'{annotation.tooltip_map[e]}s are not set up to be coded in episodes_to_code, so no plane(s) should be set up for {annotation.tooltip_map[e]}s.'}})
        return problems

    def _check_auto_coding_setup(self, strict_check) -> type_utils.ProblemDict:
        problems = self._check_auto_markers(strict_check)
        this_problems: type_utils.ProblemDict = {}
        if self.auto_code_sync_points:
            if annotation.Event.Sync_Camera not in self.episodes_to_code:
                if strict_check:
                    raise ValueError(f'The auto_code_sync_points option is configured, but {annotation.tooltip_map[annotation.Event.Sync_Camera]}s are not set to be coded in episodes_to_code. Fix episodes_to_code.')
                else:
                    this_problems['episodes_to_code'] = f'The auto_code_sync_points option is configured, but {annotation.tooltip_map[annotation.Event.Sync_Camera]}s are not set to be coded in episodes_to_code.'
                    this_problems['auto_code_sync_points'] = f'The auto_code_sync_points option is configured, but {annotation.tooltip_map[annotation.Event.Sync_Camera]}s are not set to be coded in episodes_to_code. Fix episodes_to_code or remove auto_code_sync_points setup.'
        if self.auto_code_episodes:
            for e in self.auto_code_episodes:
                if e not in self.episodes_to_code:
                    if strict_check:
                        raise ValueError(f'Coding of {annotation.tooltip_map[e]}s is configured in the auto_code_episodes option, but {annotation.tooltip_map[e]}s are not set to be coded in episodes_to_code. Fix episodes_to_code.')
                    else:
                        this_problems['episodes_to_code'] = f'Coding of {annotation.tooltip_map[e]}s is configured in the auto_code_episodes option, but {annotation.tooltip_map[e]}s are not set to be coded in episodes_to_code.'
                        this_problems['auto_code_episodes'] = f'Coding of {annotation.tooltip_map[e]}s is configured in the auto_code_episodes option, but {annotation.tooltip_map[e]}s are not set to be coded in episodes_to_code. Fix episodes_to_code or remove {annotation.tooltip_map[e]}s from the auto_code_episodes setup.'
        return type_utils.merge_problem_dicts(problems,this_problems)

    def _check_auto_markers(self, strict_check) -> type_utils.ProblemDict:
        problems: type_utils.ProblemDict = {}
        used_markers: dict[tuple[str,str]|tuple[str,annotation.Event,str],list[gt_marker.MarkerID]] = {}
        if self.auto_code_sync_points:
            if 'markers' not in self.auto_code_sync_points:
                if strict_check:
                    raise ValueError('auto_code_sync_points.markers cannot be empty or unspecified')
                else:
                    problems['auto_code_sync_points'] = {}
                    problems['auto_code_sync_points']['markers'] = 'auto_code_sync_points.markers cannot be empty or unspecified'
            else:
                missing_markers: list[gt_marker.MarkerID] = []
                for i in self.auto_code_sync_points['markers']:
                    if not any([m.id==i.m_id and m.aruco_dict_id==i.aruco_dict_id for m in self.individual_markers]):
                        if strict_check:
                            raise ValueError(f'Marker "{gt_marker.marker_ID_to_str(i)}" specified in auto_code_sync_points.markers, but unknown because not present in individual_markers')
                        else:
                            missing_markers.append(i)
                if missing_markers:
                    problems['auto_code_sync_points'] = {}
                    missing_markers = ', '.join((gt_marker.marker_ID_to_str(m) for m in missing_markers))
                    problems['auto_code_sync_points']['markers'] = f'The marker(s) {missing_markers} are not defined in individual_markers'
                used_markers[('auto_code_sync_points','markers')] = self.auto_code_sync_points['markers']
        if self.auto_code_episodes:
            for e in self.auto_code_episodes:
                for f in ['start_markers','end_markers']:
                    if f not in self.auto_code_episodes[e] or not self.auto_code_episodes[e][f]:
                        if strict_check:
                            raise ValueError(f'auto_code_episodes[{e}].{f} cannot be empty or unspecified')
                        else:
                            type_utils.merge_problem_dicts(problems, {'auto_code_episodes': {e: {f: f'auto_code_episodes[{e}].{f} cannot be empty or unspecified'}}})
                    else:
                        missing_markers: list[gt_marker.MarkerID] = []
                        for i in self.auto_code_episodes[e][f]:
                            if not any([m.id==i.m_id and m.aruco_dict_id==i.aruco_dict_id for m in self.individual_markers]):
                                if strict_check:
                                    raise ValueError(f'Marker "{gt_marker.marker_ID_to_str(i)}" specified in auto_code_episodes.[{e}].{f}, but unknown because not present in individual_markers')
                                else:
                                    missing_markers.append(i)
                        if missing_markers:
                            missing_markers = ', '.join((gt_marker.marker_ID_to_str(m) for m in missing_markers))
                            type_utils.merge_problem_dicts(problems, {'auto_code_episodes': {e: {f: f'The marker(s) {missing_markers} are not defined in individual_markers'}}})
                        used_markers[('auto_code_episodes',e,f)] = self.auto_code_episodes[e][f]
        # check if markers or marker sequences are uniquely used:
        # 1. marker used for auto_code_sync_points cannot appear anywhere else
        # 2. marker sequences used for auto_code_episodes must be unique (markers can be reused)
        # first transform marker IDs to family so we can properly detect clashes
        used_markers      : dict[tuple[str,str]|tuple[str,annotation.Event,str],list[tuple[int,int]]] = {k:[(m.m_id, aruco.dict_id_to_family[m.aruco_dict_id]) for m in used_markers[k]] for k in used_markers}
        seen_markers      : set[tuple[int,int]] = set()
        seen_markers_sets : set[tuple[tuple[int,int]]] = set()
        def _format_key(key: tuple[str,str]|tuple[str,annotation.Event,str]):
            return f'{key[0]}.{key[1]}' if len(key)==2 else f'{key[0]}[{key[1]}].{key[2]}'
        for s in used_markers:
            # first check if used markers are unique at the family level
            seen: set[tuple[int,int]] = set()
            if (duplicates := {x for x in used_markers[s] if x in seen or seen.add(x)}):
                msg = f'The markers defined for {_format_key(s)} are not unique. Please resolves the following duplicates: {utils.format_duplicate_markers_msg(duplicates)}'
                if strict_check:
                    raise ValueError(msg)
                else:
                    type_utils.merge_problem_dicts(problems, {s[0]: ({s[1]: msg} if len(s)==2 else {s[1]: {s[2]: msg}})})
            # then check if already used in another setup
            if seen_markers.intersection(used_markers[s]):
                # markers not unique, make error. Find exactly where the overlap is
                # yes, this is bad algorithmic complexity, but it only runs in failure cases
                for s2 in used_markers:
                    if s==s2 or s[0]=='auto_code_episodes' and s[0]==s2[0]:
                        # if both are different entries in 'auto_code_episodes', that is not an error
                        # for 'auto_code_episodes' we need to check marker sequences are unique, not
                        # individual entries. e.g. start is 80 81, and end is 81 80 is valid
                        continue
                    if (overlap:=set(used_markers[s2]).intersection(used_markers[s])):
                        msg = f'The following markers are encountered in the setup for both {_format_key(s)} and {_format_key(s2)}: {utils.format_duplicate_markers_msg(overlap)}. Markers cannot be used more than once, fix this collision.'
                        # emit error message
                        if strict_check:
                            raise ValueError(msg)
                        else:
                            for sx in (s,s2):
                                type_utils.merge_problem_dicts(problems, {sx[0]: ({sx[1]: msg} if len(sx)==2 else {sx[1]: {sx[2]: msg}})})
            seen_markers.update(used_markers[s])
            # check if marker sequence is already used
            if seen_markers_sets.intersection((tuple(used_markers[s]),)):
                for s2 in used_markers:
                    if s==s2:
                        continue
                    if set((tuple(used_markers[s2]),)).intersection((tuple(used_markers[s]),)):
                        msg = f'The marker sequence {utils.format_marker_sequence_msg(used_markers[s])} specified for {_format_key(s)} has already been used for {_format_key(s2)}. Markers sequences must be unique, please fix this collision.'
                        # emit error message
                        if strict_check:
                            raise ValueError(msg)
                        else:
                            for sx in (s,s2):
                                type_utils.merge_problem_dicts(problems, {sx[0]: {sx[1]: {sx[2]: msg}}})
            seen_markers_sets.add(tuple(used_markers[s]))
        return problems

    def _check_individual_markers(self, strict_check):
        problems: type_utils.ProblemDict = {}
        for m in self.individual_markers:
            problem = ''
            if m.id>=(ds:=aruco.get_dict_size(m.aruco_dict_id)):
                problem = f'dictionary {aruco.dict_id_to_str[m.aruco_dict_id]} only has {ds} markers, which means that valid IDs are 0-{ds-1}. {m.id} is thus not a valid marker for this dictionary'
            elif m.detect_only and m.size is not None:
                problem = f'size should not be set for detect only markers'
            elif not m.detect_only and (m.size is None or m.size<=0):
                problem = f'size should be set to a value larger than 0'
            elif m.marker_border_bits<1:
                problem = 'marker_border_bits must be at least 1'
            if problem:
                if strict_check:
                    raise ValueError(f'individual_markers marker {m.id} ({aruco.dict_id_to_str[m.aruco_dict_id]}): {problem}')
                else:
                    problems = type_utils.merge_problem_dicts(problems, {'individual_markers': {(m.id, aruco.dict_id_to_family[m.aruco_dict_id]): problem}})
        return problems

    def _check_sync_ref(self, strict_check):
        problems: type_utils.ProblemDict = {}
        if self.sync_ref_recording is None:
            if len(self.session_def.recordings)>1:
                problems['sync_ref_recording'] = f'sync_ref_recording must be set when sessions consist of more than one recording'
            # nothing to do
            return problems
        elif len(self.session_def.recordings)==1:
            return {'sync_ref_recording': f'sync_ref_recording must not be set when sessions consist of only one recording'}

        type_utils.merge_problem_dicts(problems, self._check_recordings([self.sync_ref_recording], 'sync_ref_recording', False))
        type_utils.merge_problem_dicts(problems, self._check_recordings(self.sync_ref_average_recordings, 'sync_average_recordings', False))
        if self.sync_ref_do_time_stretch is None:
            if strict_check:
                raise ValueError(f'sync_ref_do_time_stretch should be set in the study setup when sync_ref_recording is set')
            else:
                problems['sync_ref_do_time_stretch'] = f'sync_ref_do_time_stretch should be set when sync_ref_recording is set'
        if self.sync_ref_do_time_stretch:
            for a in ['sync_ref_stretch_which', 'sync_ref_average_recordings']:
                if getattr(self,a) is None:
                    if strict_check:
                        raise ValueError(f'{a} should be set in the study setup when sync_ref_recording is set and sync_ref_do_time_stretch is enabled')
                    else:
                        problems[a] = f'{a} should be set when sync_ref_recording is set and sync_ref_do_time_stretch is enabled'
        if self.sync_ref_average_recordings and self.sync_ref_recording in self.sync_ref_average_recordings:
            if strict_check:
                raise ValueError(f'Recording {self.sync_ref_recording} is the reference recording for sync, should not be specified in sync_average_recordings')
            else:
                problems['sync_ref_average_recordings'] = f'Recording {self.sync_ref_recording} is the reference recording for sync, cannot be specified in sync_average_recordings'
        if annotation.Event.Sync_Camera not in self.episodes_to_code:
            if strict_check:
                raise ValueError('when sync_ref_recording is set, coding of camera sync points should be set up in episodes_to_code')
            else:
                problems['episodes_to_code'] = f'if sync_ref_recording is set, {annotation.tooltip_map[annotation.Event.Sync_Camera]}s should be set up to be coded'
                type_utils.merge_problem_dicts(problems, {'sync_ref_recording': f'sync_ref_recording is set, but {annotation.tooltip_map[annotation.Event.Sync_Camera]}s are not set up to be coded in episodes_to_code'})
        return problems

    def _check_et_sync_method(self, strict_check) -> type_utils.ProblemDict:
        problems: type_utils.ProblemDict = {}
        cam_mov_possible_values = typing.get_args(study_parameter_types['get_cam_movement_for_et_sync_method'])
        if self.get_cam_movement_for_et_sync_method not in cam_mov_possible_values:
            values = list(cam_mov_possible_values)
            values.remove('')
            values_str = '"' + '", "'.join(values) + '"'
            temp = values_str.partition(f'"{values[-1]}"')
            values_str = ('' if len(values)==1 else ', ') + temp[0] + 'or ' + temp[1]
            if strict_check:
                raise ValueError(f'get_cam_movement_for_et_sync_method parameter should be an empty string{values_str}')
            else:
                problems['get_cam_movement_for_et_sync_method'] = f'get_cam_movement_for_et_sync_method parameter should be an empty string{values_str}'

        if self.get_cam_movement_for_et_sync_method not in ['plane', 'function']:
            # nothing to do
            return problems
        if annotation.Event.Sync_ET_Data not in self.episodes_to_code:
            if strict_check:
                raise ValueError(f'if get_cam_movement_for_et_sync_method is set to "plane" or "function", {annotation.tooltip_map[annotation.Event.Sync_ET_Data]}s should be set up to be coded in episodes_to_code')
            else:
                problems['episodes_to_code'] = f'if get_cam_movement_for_et_sync_method is set to "plane" or "function", {annotation.tooltip_map[annotation.Event.Sync_ET_Data]}s should be set up to be coded'
                problems['get_cam_movement_for_et_sync_method'] = f'get_cam_movement_for_et_sync_method is set to "{self.get_cam_movement_for_et_sync_method}", but {annotation.tooltip_map[annotation.Event.Sync_ET_Data]}s are not set up to be coded in episodes_to_code'
        if self.get_cam_movement_for_et_sync_method=='function':
            if strict_check:
                if not self.get_cam_movement_for_et_sync_function or not all([x in self.get_cam_movement_for_et_sync_function for x in ["module_or_file","function","parameters"]]):
                    raise ValueError('if get_cam_movement_for_et_sync_method is set to "function", get_cam_movement_for_et_sync_function should be a dict specifying "module_or_file", "function", and "parameters"')
            else:
                t = gt_utils.unpack_none_union(study_parameter_types['get_cam_movement_for_et_sync_function'])[0]
                keys = t.__required_keys__|t.__optional_keys__
                this_problems = {k:f'{k} should be set when get_cam_movement_for_et_sync_function is set to "function"' for k in keys if not self.get_cam_movement_for_et_sync_function or (k not in t._field_defaults and (k not in self.get_cam_movement_for_et_sync_function or not self.get_cam_movement_for_et_sync_function[k]))}
                if this_problems:
                    problems['get_cam_movement_for_et_sync_function'] = this_problems
        elif self.get_cam_movement_for_et_sync_method=='plane':
            if annotation.Event.Sync_ET_Data not in self.planes_per_episode:
                if strict_check:
                    raise ValueError(f'if get_cam_movement_for_et_sync_method is set to "plane", a plane should be set up to be used for processing {annotation.tooltip_map[annotation.Event.Sync_ET_Data]}s in planes_per_episode')
                else:
                    problems['planes_per_episode'] = f'if get_cam_movement_for_et_sync_method is set to "plane", a plane should be set up to be used for processing {annotation.tooltip_map[annotation.Event.Sync_ET_Data]}s'
                    type_utils.merge_problem_dicts(problems, {'get_cam_movement_for_et_sync_method': f'get_cam_movement_for_et_sync_method is set to "plane", but no plane specified for syncing eye tracker data to the scene cam (i.e., for {annotation.tooltip_map[annotation.Event.Sync_ET_Data]}s) in planes_per_episode'})
        return problems

    def _check_make_video(self, strict_check) -> type_utils.ProblemDict:
        problems = self._check_recordings(self.mapped_video_make_which, 'mapped_video_make_which', strict_check)
        type_utils.merge_problem_dicts(problems,
                   self._check_recordings(self.mapped_video_recording_colors, 'mapped_video_recording_colors', strict_check))
        type_utils.merge_problem_dicts(problems,
                   self._check_recordings(self.mapped_video_show_gaze_on_plane_in_which, 'mapped_video_show_gaze_on_plane_in_which', strict_check))
        type_utils.merge_problem_dicts(problems,
                   self._check_recordings(self.mapped_video_show_camera_in_which, 'mapped_video_show_camera_in_which', strict_check))
        type_utils.merge_problem_dicts(problems,
                   self._check_recordings(self.mapped_video_show_gaze_vec_in_which, 'mapped_video_show_gaze_vec_in_which', strict_check))
        if self.mapped_video_make_which:
            # check have colors for all eye tracker recordings
            all_recs = {r.name for r in self.session_def.recordings if r.type==session.RecordingType.Eye_Tracker}
            if self.mapped_video_recording_colors:
                missing = list(all_recs-set(self.mapped_video_recording_colors.keys()))
            else:
                missing = list(all_recs)
            if missing:
                msg = f'Colors need to be defined for all eye tracker recordings. Missing for {missing[0] if len(missing)==1 else missing}'
                if strict_check:
                    raise ValueError(msg)
                else:
                    type_utils.merge_problem_dicts(problems,{'mapped_video_recording_colors': msg})
        return problems

    def field_problems(self) -> type_utils.ProblemDict:
        problems: type_utils.ProblemDict = {}
        type_utils.merge_problem_dicts(problems, self._check_session_def(False))
        type_utils.merge_problem_dicts(problems, self._check_planes_per_episode(False))
        type_utils.merge_problem_dicts(problems, self._check_episodes_to_code(False))
        type_utils.merge_problem_dicts(problems, self._check_auto_coding_setup(False))
        type_utils.merge_problem_dicts(problems, self._check_individual_markers(False))
        type_utils.merge_problem_dicts(problems, self._check_sync_ref(False))
        type_utils.merge_problem_dicts(problems, self._check_et_sync_method(False))
        type_utils.merge_problem_dicts(problems, self._check_make_video(False))
        return problems

    def store_as_json(self, path: str|pathlib.Path|None=None):
        if not path:
            path = guess_config_dir(self.working_directory)
        path = pathlib.Path(path)
        # this stores only the planes_per_episode variable to json, rest is read from other files
        # instead to remain flexible and make it easy for users to rename, etc
        f_path = path
        if f_path.is_dir():
            f_path /= self.default_json_file_name
        else:
            path = f_path.parent
            # prep for dump to file
        to_dump = {k:copy.deepcopy(getattr(self,k)) for k in vars(self) if not k.startswith('_') and k not in ['session_def','planes','working_directory']}    # session_def and planes will be populated from contents in the provided folder, and working_directory as the provided path
        to_dump['planes_per_episode'] = [(k, to_dump['planes_per_episode'][k]) for k in to_dump['planes_per_episode']]   # pack as list of tuples for storage
        # filter out defaulted
        to_dump = {k:to_dump[k] for k in to_dump if k not in study_defaults or study_defaults[k]!=to_dump[k]}
        # also filter out defaults in some subfields
        typed_dict_default_fields = [k for k in to_dump if typed_dict_defaults.is_typeddictdefault(gt_utils.unpack_none_union(study_parameter_types[k])[0])]
        for k in typed_dict_default_fields:
            to_dump[k] = {kk:to_dump[k][kk] for kk in to_dump[k] if kk not in to_dump[k]._field_defaults or to_dump[k]._field_defaults[kk]!=to_dump[k][kk]}
            if not to_dump[k] and k in study_defaults and study_defaults[k] is not None:
                to_dump.pop(k)
        if 'auto_code_episodes' in to_dump:
            for e in to_dump['auto_code_episodes']:
                to_dump['auto_code_episodes'][e] = {kk:to_dump['auto_code_episodes'][e][kk] for kk in to_dump['auto_code_episodes'][e] if kk not in to_dump['auto_code_episodes'][e]._field_defaults or to_dump['auto_code_episodes'][e]._field_defaults[kk]!=to_dump['auto_code_episodes'][e][kk]}
            # pack as list of tuples for storage
            to_dump['auto_code_episodes'] = [(e, to_dump['auto_code_episodes'][e]) for e in to_dump['auto_code_episodes']]   # pack as list of tuples for storage
        # dump to file
        json.dump(to_dump, f_path)
        # this doesn't store any files itself, but triggers the contained info to be stored
        self.session_def.store_as_json(path)
        for p in self.planes:
            p_dir = path / p.name
            if not p_dir.is_dir():
                p_dir.mkdir()
            p.store_as_json(p_dir)

    @staticmethod
    def get_empty(path: str | pathlib.Path) -> 'Study':
        # returns a minimally set up config, with every required argument set empty, and the rest default-initialized
        return Study(
            session.SessionDefinition(),
            [],{},set(),[],path,
            strict_check=False
        )

    @staticmethod
    def load_from_json(path: str | pathlib.Path, strict_check: bool=True) -> 'Study':
        path = pathlib.Path(path)
        if path.is_dir():
            d_path = path / Study.default_json_file_name
        else:
            d_path = path
        # get kwds
        kwds = json.load(d_path)
        # fix up 'video_' parameters for backwards compatibility
        for k in list(kwds.keys()):
            if k.startswith('video_'):
                kwds[f'mapped_{k}'] = kwds.pop(k)
        # stored as list of tuples (with enum values as keys), unpack
        kwds['planes_per_episode'] = {annotation.Event(k):v for k,v in kwds['planes_per_episode']}
        if 'auto_code_episodes' in kwds:
            kwds['auto_code_episodes'] = {annotation.Event(k):v for k,v in kwds['auto_code_episodes']}
        # help with enum roundtrip
        kwds['episodes_to_code'] = {annotation.Event(e) for e in kwds['episodes_to_code']}
        if 'validate_dq_types' in kwds:
            kwds['validate_dq_types']= {DataQualityType(d) for d in kwds['validate_dq_types']}
        if 'mapped_video_which_gaze_type_on_plane' in kwds:
            kwds['mapped_video_which_gaze_type_on_plane'] = gaze_worldref.Type(kwds['mapped_video_which_gaze_type_on_plane'])
        # backward compatibility: ensure value are stored in sets, not lists
        kwds['planes_per_episode'] = {k:set(v) for k,v in kwds['planes_per_episode'].items()}
        if 'sync_ref_average_recordings' in kwds:
            kwds['sync_ref_average_recordings'] = set(kwds['sync_ref_average_recordings'])
        if 'mapped_video_make_which' in kwds:
            kwds['mapped_video_make_which'] = set(kwds['mapped_video_make_which'])
        # backwards compatibility, help with named tuple roundtrip
        for k in ('overlay_video_gaze_vid_pos_color','overlay_video_gaze_world_pos_color','mapped_video_projected_vidPos_color','mapped_video_projected_world_pos_color','mapped_video_projected_left_ray_color','mapped_video_projected_right_ray_color','mapped_video_projected_average_ray_color'):
            if k in kwds and kwds[k] is not None and not isinstance(kwds[k],RgbColor):
                kwds[k] = RgbColor(*kwds[k])
        if 'mapped_video_recording_colors' in kwds and any((not isinstance(kwds['mapped_video_recording_colors'][k],RgbColor) for k in kwds['mapped_video_recording_colors'])):
            kwds['mapped_video_recording_colors'] = {k: RgbColor(*kwds['mapped_video_recording_colors'][k]) for k in kwds['mapped_video_recording_colors']}
        # backwards compatibility, rename 'auto_code_trial_episodes'
        if 'auto_code_trial_episodes' in kwds:
            kwds['auto_code_episodes'] = {annotation.Event.Trial: kwds.pop('auto_code_trial_episodes')}
        # backwards compatibility, upgrade markers to markerIDs if they're bare ints
        if 'auto_code_sync_points' in kwds and kwds['auto_code_sync_points'] is not None and 'markers' in kwds['auto_code_sync_points']:
            markers = kwds['auto_code_sync_points']['markers']
            kwds['auto_code_sync_points']['markers'] = set()
            for m in markers:
                if not isinstance(m,int):
                    kwds['auto_code_sync_points']['markers'].add(m)
                    continue
                # find corresponding marker in individual_markers
                im = [im for im in kwds['individual_markers'] if im.id==m]
                if not im:
                    # incorrect setup, referring to a non-existing individual marker. ignore
                    continue
                else:
                    kwds['auto_code_sync_points']['markers'].add(gt_marker.MarkerID(im[0].id, im[0].aruco_dict_id))
        if 'auto_code_episodes' in kwds:
            for e in kwds['auto_code_episodes']:
                for f in ('start_markers','end_markers'):
                    if not f in kwds['auto_code_episodes'][e]:
                        continue
                    markers = kwds['auto_code_episodes'][e][f]
                    kwds['auto_code_episodes'][e][f] = []
                    for m in markers:
                        if not isinstance(m,int):
                            kwds['auto_code_episodes'][e][f].append(m)
                            continue
                        # find corresponding marker in individual_markers
                        im = [im for im in kwds['individual_markers'] if im.id==m]
                        if not im:
                            # incorrect setup, referring to a non-existing individual marker. ignore
                            continue
                        else:
                            kwds['auto_code_episodes'][e][f].append(gt_marker.MarkerID(im[0].id, im[0].aruco_dict_id))
        # backwards compatibility for mapped video marker and axes settings
        kwds.pop('mapped_video_process_individual_markers_for_all_frames',None) # setting doesn't exist anymore
        if 'mapped_video_show_detected_markers' in kwds:
            mapped_video_show_detected_markers = kwds.pop('mapped_video_show_detected_markers')
            kwds['mapped_video_plane_marker_color'] = study_defaults['mapped_video_plane_marker_color'] if mapped_video_show_detected_markers else None
            kwds['mapped_video_recovered_plane_marker_color'] = study_defaults['mapped_video_recovered_plane_marker_color'] if mapped_video_show_detected_markers else None
        if 'mapped_video_show_plane_axes' in kwds or 'mapped_video_show_board_axes' in kwds:
            mapped_video_show_plane_axes = kwds.pop('mapped_video_show_plane_axes') if 'mapped_video_show_plane_axes' in kwds else kwds.pop('mapped_video_show_board_axes')
            kwds['mapped_video_plane_axis_arm_length'] = study_defaults['mapped_video_plane_axis_arm_length'] if mapped_video_show_plane_axes else None
        if 'mapped_video_show_individual_marker_axes' in kwds:
            mapped_video_show_individual_marker_axes = kwds.pop('mapped_video_show_individual_marker_axes')
            kwds['mapped_video_individual_marker_axis_arm_length'] = study_defaults['mapped_video_individual_marker_axis_arm_length'] if mapped_video_show_individual_marker_axes else None
        if 'mapped_video_show_unexpected_markers' in kwds:
            mapped_video_show_unexpected_markers = kwds.pop('mapped_video_show_unexpected_markers')
            kwds['mapped_video_unexpected_marker_color'] = study_defaults['mapped_video_unexpected_marker_color'] if mapped_video_show_unexpected_markers else None
        if 'mapped_video_show_rejected_markers' in kwds:
            mapped_video_show_rejected_markers = kwds.pop('mapped_video_show_rejected_markers')
            kwds['mapped_video_rejected_marker_color'] = study_defaults['mapped_video_rejected_marker_color'] if mapped_video_show_rejected_markers else None

        # get session def
        s_path = path / 'session_def.json'
        if not s_path.is_file():
            return None
        sess_def = session.SessionDefinition.load_from_json(s_path)

        # get planes
        planes: list[plane.Definition] = []
        for p_dir in path.iterdir():
            if not p_dir.is_dir():
                continue
            p_file = p_dir / plane.Definition.default_json_file_name
            if not p_file.is_file():
                continue
            planes.append(plane.Definition.load_from_json(p_file))

        return Study(sess_def, planes, working_directory=path.parent, **kwds, strict_check=strict_check)

# get defaults for default argument of Study constructor
_params = inspect.signature(Study.__init__).parameters
study_defaults = {k:d for k in _params if (d:=_params[k].default)!=inspect._empty}
study_parameter_types = {k:_params[k].annotation for k in _params if k not in ['self','strict_check']}
def _get_gv_data_quality_type_doc(dq: DataQualityType):
    t,doc = get_DataQualityType_explanation(dq)
    return (dq, type_utils.GUIDocInfo(t,doc))
def _get_annotation_event_doc(a: annotation.Event, children: dict = None):
    t = annotation.tooltip_map[a]
    doc = {
        annotation.Event.Trial: 'Denotes an episode for which to map gaze to plane(s). This determines for which segments there will be gaze data when running the Export Trials.',
        annotation.Event.Validate: 'Denotes an episode during which a participant looked at a validation poster, to be used to run glassesValidator to compute data quality of the gaze data.',
        annotation.Event.Sync_Camera: 'Time point (frame from video) when a synchronization event happened, used for synchronizing different recordings.',
        annotation.Event.Sync_ET_Data: 'Episode to be used for synchronization of eye tracker data to scene camera (e.g. using VOR).'
    }.get(a)
    if children is None:
        children = {}
    return (a, type_utils.GUIDocInfo(t,doc, children))
_gaze_type_doc = {
    gaze_worldref.Type.Scene_Video_Position : type_utils.GUIDocInfo('Gaze position on scene video', 'Gaze position on the scene video, projected to the plane.'),
    gaze_worldref.Type.World_3D_Point       : type_utils.GUIDocInfo('3D gaze position', '3D gaze position in the world provided by the eye tracker, projected to the plane.'),
    gaze_worldref.Type.Left_Eye_Gaze_Vector : type_utils.GUIDocInfo('Left eye gaze vector', 'Projection of the left eye\'s gaze vector to the plane.'),
    gaze_worldref.Type.Right_Eye_Gaze_Vector: type_utils.GUIDocInfo('Right eye gaze vector', 'Projection of the right eye\'s gaze vector to the plane.'),
    gaze_worldref.Type.Average_Gaze_Vector  : type_utils.GUIDocInfo('Average of gaze vectors', 'Average of the projections of the left and right eyes\' gaze vectors to the plane.'),
}
study_parameter_doc = {
    'planes_per_episode': type_utils.GUIDocInfo('Planes per episode', 'For each episode that is enabled to be coded in the project, sets which planes will be looked for and gaze mapped to during the episode.',dict([_get_annotation_event_doc(a) for a in annotation.Event])),
    'episodes_to_code': type_utils.GUIDocInfo('Episodes to code', 'Sets which episodes can be coded for this project.',{
        None: # None indicates the doc specification applies to the contained values
            dict([_get_annotation_event_doc(a) for a in annotation.Event])
    }),
    'import_do_copy_video': type_utils.GUIDocInfo('Copy video during import?', 'If not enabled, the scene video of an eye tracker recording, or the video of an external camera is not copied to the gazeMapper recording directory during import. Instead, the video will be loaded from the recording\'s source directory (so do not move it). Ignored when the video must be transcoded to be processed with gazeMapper.'),
    'import_source_dir_as_relative_path': type_utils.GUIDocInfo('Store source directory as relative path?', 'Specifies whether the path to the source directory stored in the recording info file is an absolute path (this option is not enabled) or a relative path (enabled). If a relative path is used, the imported recording and the source directory can be moved to another location, and the source directory can still be found as long as the relative path (e.g., one folder up and in the directory "original recordings": "../original recordings") doesn\'t change.'),
    'import_known_custom_eye_trackers': type_utils.GUIDocInfo('Registered custom eye trackers', 'gazeMapper allows importing generic eye trackers for which no specific support is implemented, if their recording data is preprocessed to conform to glassesTools\' generic data format. Here you can define specific known generic eye tracker names that you may import.'),
    'overlay_video_gaze_vid_pos_color': type_utils.GUIDocInfo('Gaze overlay video: Color for gaze position on video', 'Color used for drawing the recorded gaze position on the scene video.'),
    'overlay_video_gaze_world_pos_color': type_utils.GUIDocInfo('Gaze overlay video: Color for 3D gaze position', 'Color used for drawing the recorded 3D gaze position in the world. Not drawn if value is not set.'),
    'overlay_video_gaze_vid_pos_radius': type_utils.GUIDocInfo('Gaze overlay video: Radius for gaze position on video', 'Radius of circle used for drawing the recorded gaze position on the scene video.'),
    'overlay_video_gaze_world_pos_radius': type_utils.GUIDocInfo('Gaze overlay video: Radius for 3D gaze position', 'Radius of circle used for drawing the recorded 3D gaze position in the world.'),
    'overlay_video_gaze_vid_pos_thickness': type_utils.GUIDocInfo('Gaze overlay video: Thickness for gaze position on video', 'Line thickness of circle used for drawing the recorded gaze position on the scene video.'),
    'overlay_video_gaze_world_pos_thickness': type_utils.GUIDocInfo('Gaze overlay video: Thickness for 3D gaze position', 'Line thickness of circle used for drawing the recorded 3D gaze position in the world.'),
    'sync_ref_recording': type_utils.GUIDocInfo('Synchronization: Reference recording', 'If there are multiple recordings, sets to which recording all other recordings will be synchronized.'),
    'sync_ref_do_time_stretch': type_utils.GUIDocInfo('Synchronization: Do time stretch?', 'If enabled, multiple sync points are used to calculate a time stretch factor to compensate for clock drift when synchronizing multiple recordings.'),
    'sync_ref_stretch_which': type_utils.GUIDocInfo('Synchronization: Stretch which recording', 'Which recording(s) should be corrected for clock drift if "Synchronization: Do time stretch?" is enabled.',{
        None: {     # indicates the doc specification applies to the contained values
            'ref': type_utils.GUIDocInfo('Reference recording', 'The time signal of the reference recording is stretched to compensate for clock drift.'),
            'other': type_utils.GUIDocInfo('Other recording(s)', 'The time signal of the other recording(s) is stretched to compensate for clock drift.')
        }
    }),
    'sync_ref_average_recordings': type_utils.GUIDocInfo('Synchronization: Average recordings?', 'Whether to average the clock drifts for multiple recordings if "Synchronization: Do time stretch?" is enabled.'),
    'get_cam_movement_for_et_sync_method': type_utils.GUIDocInfo('Gaze data synchronization: Method to get camera movement', 'Method used to derive the head motion for synchronizing eye tracker data and scene camera.',{
        None: {     # indicates the doc specification applies to the contained values
            '': type_utils.GUIDocInfo('None', 'No gaze data synchronization'),
            'plane': type_utils.GUIDocInfo('Plane', 'Head movement is represented by the position of the origin of the plane in the scene camera video (the plane that is set up to be used for "Sync ET Data" episodes), as extracted through pose estimation or homography using a gazeMapper plane.'),
            'function': type_utils.GUIDocInfo('Function', 'A user-specified function (configured using the "Gaze data synchronization: Function for camera movement" setting) will be called for each frame of the scene video in a "Sync ET Data" episode and is expected to return the location of the target the participant was looking at.')
        }
    }),
    'get_cam_movement_for_et_sync_function': type_utils.GUIDocInfo('Gaze data synchronization: Function for camera movement', 'Setup for function to use for deriving the head motion when synchronizing eye tracker data and scene camera if "Gaze data synchronization: Method to get camera movement" is set to "function".',{
        'module_or_file': type_utils.GUIDocInfo('Module or file', 'Importable module or file (can be a full path) that contains the function to run.'),
        'function': type_utils.GUIDocInfo('Function', 'Name of the function to run.'),
        'parameters': type_utils.GUIDocInfo('Parameters', 'Set of parameters and values to pass to the function. The frame to process (np.ndarray) is the first (positional) input passed to the function, and should not be specified in this set.'),
    }),
    'sync_et_to_cam_use_average': type_utils.GUIDocInfo('Gaze data synchronization: Use average?', 'Whether to use the average offset of multiple sync episodes. If not enabled, the offset for the first sync episode is used, the rest are ignored.'),
    'auto_code_sync_points': type_utils.GUIDocInfo('Automated coding of synchronization points','Setup for automatic coding of synchronization timepoints.',{
        'markers': type_utils.GUIDocInfo('Marker(s)', 'Set of marker IDs whose appearance indicates a synchronization timepoint.'),
        'max_gap_duration': type_utils.GUIDocInfo('Maximum gap duration', 'Maximum gap (number of frames) in marker detections that will be filled in (ignored).'),
        'min_duration': type_utils.GUIDocInfo('Minimum duration', 'Minimum duration (number of frames) that a marker should be detected. Shorter runs are removed.')
    }),
    'auto_code_episodes': type_utils.GUIDocInfo('Automated coding of episodes','Setup for automatic coding of episodes.',{
        None: # None indicates the doc specification applies to the contained values
            dict([_get_annotation_event_doc(a,{
                'start_markers': type_utils.GUIDocInfo('Start marker(s)', 'A single marker ID or a sequence of marker IDs that indicate the start of an episode.'),
                'end_markers': type_utils.GUIDocInfo('End marker(s)', 'A single marker ID or a sequence of marker IDs that indicate the end of an episode.'),
                'max_gap_duration': type_utils.GUIDocInfo('Maximum gap duration', 'Maximum gap (number of frames) in marker detections that will be filled in (ignored).'),
                'max_intermarker_gap_duration': type_utils.GUIDocInfo('Maximum intermarker gap duration', 'Maximum gap (number of frames) that is allowed between the detection of two markers in a sequence.'),
                'min_duration': type_utils.GUIDocInfo('Minimum duration', 'Minimum duration (number of frames) that a marker should be detected. Shorter runs are removed.')
                }
            ) for a in annotation.Event if a!=annotation.Event.Sync_Camera])
    }),
    'export_output3D': type_utils.GUIDocInfo('Mapped data export: include 3D fields', 'Determines whether gaze positions on the plane in the scene camera reference frame are exported when invoking the Export Trials action.'),
    'export_output2D': type_utils.GUIDocInfo('Mapped data export: include 2D fields', 'Determines whether gaze positions on the plane in the plane\'s reference frame are exported when invoking the Export Trials action.'),
    'export_only_code_marker_presence': type_utils.GUIDocInfo('Mapped data export: only include marker presence?', 'If enabled, for each marker only a single column is added to the export created by the Export Trials action, indicating whether the given marker was detected or not on a given frame. If not enabled, marker pose information is included in the export.'),
    'validate_do_global_shift': type_utils.GUIDocInfo('glassesValidator: Apply global shift?', 'If enabled, for each validation interval the median position will be removed from the gaze data and the mean from the targets, removing any overall shift of the data. This improves the matching of fixations to targets when there is a significant overall offset in the data. It may fail (backfire) if there are data samples far outside the range of the validation targets, or if there is no data for some targets.'),
    'validate_max_dist_fac': type_utils.GUIDocInfo('glassesValidator: Maximum distance factor', 'Factor for determining distance limit when assigning fixation points to validation targets. If for a given target the closest fixation point is further away than <factor>*[minimum intertarget distance], then no fixation point will be assigned to this target, i.e., it will not be matched to any fixation point. Set to a large value to essentially disable.'),
    'validate_dq_types': type_utils.GUIDocInfo('glassesValidator: Data quality types', 'Selects the types of data quality you would like to calculate for each of the recordings. When none are selected, a good default is used for each recording. When none of the selected types is available, depending on the `validate_allow_dq_fallback` setting, either an error is thrown or that same default is used instead. Whether a data quality type is available depends on what type of gaze information is available for a recording, as well as whether the camera is calibrated.',{
        None: # None indicates the doc specification applies to the contained values
            dict([_get_gv_data_quality_type_doc(dq) for dq in DataQualityType])
    }),
    'validate_allow_dq_fallback': type_utils.GUIDocInfo('glassesValidator: Allow fallback data quality type?', 'If not enabled, an error is raised when the data quality type(s) indicated in "glassesValidator: Data quality types" are not available. If enabled, a sensible default other data type will be used instead. Does not apply if the "glassesValidator: Data quality types" is not set.'),
    'validate_include_data_loss': type_utils.GUIDocInfo('glassesValidator: Include data loss?', 'If enabled, the data quality report will include data loss during the episode selected for each target on the validation poster. This is NOT the data loss of the whole recording and thus not what you want to report in your paper.'),
    'validate_I2MC_settings': type_utils.GUIDocInfo('glassesValidator: I2MC settings','Settings for the I2MC fixation classifier used as part of determining the fixation that are assigned to validation targets. Settings that are "<not set>" will be determined based on the provided eye tracking data.',{
        'freq': type_utils.GUIDocInfo('Sampling frequency', 'Sampling frequency of the eye tracking data.'),
        'windowtimeInterp': type_utils.GUIDocInfo('Maximum gap duration for interpolation', 'Maximum duration (s) of gap in the data that is interpolated.'),
        'edgeSampInterp': type_utils.GUIDocInfo('# Edge samples','Amount of data (number of samples) at edges needed for interpolation.'),
        'maxdisp': type_utils.GUIDocInfo('Maximum dispersion', 'Maximum distance (mm) between the two edges of a gap below which the missing data is interpolated.'),
        'windowtime': type_utils.GUIDocInfo('Moving window duration','Length of the moving window (s) used by I2MC to calculate 2-means clustering when processing the data.'),
        'steptime': type_utils.GUIDocInfo('Moving window step', 'Step size (s) by which the moving window is moved.'),
        'downsamples': type_utils.GUIDocInfo('Downsample factors', 'Set of integer decimation factors used to downsample the gaze data as part of I2MC processing.'),
        'downsampFilter': type_utils.GUIDocInfo('Apply Chebyshev filter?', 'If enabled, a Chebyshev low-pass filter is applied when downsampling.'),
        'chebyOrder': type_utils.GUIDocInfo('Chebyshev filter order','Order of the Chebyshev low-pass filter.'),
        'maxerrors': type_utils.GUIDocInfo('Maximum # errors', 'Maximum number of errors before processing of a trial is aborted.'),
        'cutoffstd': type_utils.GUIDocInfo('Fixation cutoff factor', 'Number of standard deviations above mean k-means weights that will be used as fixation cutoff.'),
        'onoffsetThresh': type_utils.GUIDocInfo('Onset/offset Threshold', 'Number of MAD away from median fixation duration. Will be used to walk forward at fixation starts and backward at fixation ends to refine their placement and stop algorithm from eating into saccades.'),
        'maxMergeDist': type_utils.GUIDocInfo('Maximum merging distance', 'Maximum Euclidean distance (mm) between fixations for merging to be possible.'),
        'maxMergeTime': type_utils.GUIDocInfo('Maximum gap duration for merging', 'Maximum time (ms) between fixations for merging to be possible.'),
        'minFixDur': type_utils.GUIDocInfo('Minimum fixation duration', 'Minimum fixation duration (ms) after merging, fixations with shorter duration are removed from output.'),
    }),
    'validate_dynamic_skip_first_duration': type_utils.GUIDocInfo('glassesValidator: Dynamic skip first duration', 'For a glassesValidator plane that is marked as dynamic (i.e. for a validation procedure using the PsychoPy script), how many seconds of data to not use from the beginning of each target interval.'),
    'validate_dynamic_max_gap_duration': type_utils.GUIDocInfo('glassesValidator: Dynamic, maximum gap duration', 'For a glassesValidator plane that is marked as dynamic (i.e. for a validation procedure using the PsychoPy script), maximum gap (number of frames) in marker detections that will be filled in (ignored).'),
    'validate_dynamic_min_duration': type_utils.GUIDocInfo('glassesValidator: Dynamic, minimum duration', 'Minimum duration (number of frames) that a marker should be detected. Shorter runs are removed.'),
    'mapped_video_make_which': type_utils.GUIDocInfo('Mapped video: Which recordings', 'Indicates one or multiple recordings for which to make videos of the eye tracker scene camera or external camera (synchronized to one of the recordings if there are multiple) showing detected plane origins, detected individual markers and gaze from any other recordings eye tracker recordings. Also shown for eye tracker recordings are gaze on the scene video from the eye tracker, gaze projected to the detected planes. Each only if available, and enabled in the below video generation settings.'),
    'mapped_video_recording_colors': type_utils.GUIDocInfo('Mapped video: Recording colors', 'Colors used for drawing each recording\'s gaze point, scene camera and gaze vector (depending on settings).'),
    'mapped_video_projected_vidPos_color': type_utils.GUIDocInfo('Mapped video: Color for gaze position on plane', 'Color used for drawing the recorded gaze position on the scene video transformed to the plane. Not drawn if value is not set.'),
    'mapped_video_projected_world_pos_color': type_utils.GUIDocInfo('Mapped video: Color for 3D gaze position on plane', 'Color used for drawing the projection on a plane of the recorded 3D gaze position in the world. Not drawn if value is not set.'),
    'mapped_video_projected_left_ray_color': type_utils.GUIDocInfo('Mapped video: Color for left eye gaze vector projected to plane', 'Color used for drawing the projection to a plane of the recorded left eye\'s gaze vector. Not drawn if value is not set.'),
    'mapped_video_projected_right_ray_color': type_utils.GUIDocInfo('Mapped video: Color for right eye gaze vector projected to plane', 'Color used for drawing the projection to a plane of the recorded right eye\'s gaze vector. Not drawn if value is not set.'),
    'mapped_video_projected_average_ray_color': type_utils.GUIDocInfo('Mapped video: Color for average of gaze vectors projected to plane', 'Color used for drawing the average projection to a plane of the recorded left and right eyes\' gaze vectors. Not drawn if value is not set.'),
    'mapped_video_process_planes_for_all_frames': type_utils.GUIDocInfo('Mapped video: Process all planes for all frames?', 'If enabled, shows detection results for all planes for all frames. If not enabled, detection of each plane is only shown during the episode(s) to which it is assigned.'),
    'mapped_video_process_annotations_for_all_recordings': type_utils.GUIDocInfo('Mapped video: Process all annotations for all recordings?', 'Episode annotations are shown in a bar on the bottom of the screen. If enabled, annotations for not only the recording for which the video is made, but also for the other recordings are shown in this bar.'),
    'mapped_video_plane_marker_color': type_utils.GUIDocInfo('Mapped video: Plane marker color', 'Color used for drawing the detected markers belonging to planes. Not drawn if value is not set.'),
    'mapped_video_recovered_plane_marker_color': type_utils.GUIDocInfo('Mapped video: Plane marker color (recovered)', 'Color used for drawing the detected markers belonging to planes that were identified during a second stage in the ArUco pipeline where markers are found based on their expected position extrapolated from identified nearby markers on the plane. Not drawn if value is not set.'),
    'mapped_video_plane_axis_arm_length': type_utils.GUIDocInfo('Mapped video: Plane axis arm length','Length used for drawing axes that indicate the orientation of a detected plane. In the same unit as the plane (usually mm). Not drawn if value is not set.'),
    'mapped_video_individual_marker_color': type_utils.GUIDocInfo('Mapped video: Individual marker color', 'Color used for drawing the detected individual markers. Not drawn if value is not set.'),
    'mapped_video_individual_marker_axis_arm_length': type_utils.GUIDocInfo('Mapped video: Individual marker axis arm length','Length used for drawing axes that indicate the orientation of a detected individual marker (if its size is set). In the same unit as the markers (usually mm). Not drawn if value is not set.'),
    'mapped_video_unexpected_marker_color': type_utils.GUIDocInfo('Mapped video: Unexpected marker color', 'Color used for drawing the detected markers belonging to planes. Not drawn if value is not set.'),
    'mapped_video_rejected_marker_color': type_utils.GUIDocInfo('Mapped video: Rejected marker color', 'Color used for drawing marker candidates that were rejected at some stage in the ArUco pipeline. For (setup) debug purposes. Not drawn if value is not set. (255,0,0) may be a good color.'),
    'mapped_video_show_sync_func_output': type_utils.GUIDocInfo('Mapped video: Show sync function output?', 'If enabled, draw the output of the function on the output video. Applies if the "Gaze data synchronization: Function for camera movement" setting is set to "function".'),
    'mapped_video_show_gaze_on_plane_in_which': type_utils.GUIDocInfo('Mapped video: Videos on which to draw gaze projected to plane', 'For the listed recordings, gaze projected to the plane (both the gaze point on the scene video, and the left and right eyes\' gaze vectors if available) is drawn on the video for eye tracker recordings. If there are multiple recordings in a session, gaze positions for the other recordings (if available) will also be drawn on the indicated video(s).'),
    'mapped_video_show_gaze_vec_in_which': type_utils.GUIDocInfo('Mapped video: Videos on which to draw gaze vectors(s)', 'For the listed recordings, a line is drawn between the positions of the cameras of other recordings and the gaze position of these recordings projected to the plane in the generated video. Only for eye tracker recordings.'),
    'mapped_video_show_camera_in_which': type_utils.GUIDocInfo('Mapped video: Videos on which to draw camera position(s)', 'For the listed recordings, the position of the cameras of other recordings is marked in the generated video.'),
    'mapped_video_which_gaze_type_on_plane': type_utils.GUIDocInfo('Mapped video: Which gaze on plane to show?', 'Sets which gaze-on-plane (e.g. from gaze position on the scene video or from gaze vectors projected to the plane) is used for the gaze positions shown in the generated videos.', _gaze_type_doc),
    'mapped_video_which_gaze_type_on_plane_allow_fallback': type_utils.GUIDocInfo('Mapped video: Allow fallback to showing gaze on plane based on scene video gaze?', 'Sets if it is allowed to fall back to using the projection of the gaze position on the plane derived from the gaze position on the video if the gaze-on-plane type specified in the "Mapped video: Which gaze on plane to show?" setting is not available.'),
    'mapped_video_gaze_to_plane_margin': type_utils.GUIDocInfo('Mapped video: Gaze position margin','Gaze position more than this factor outside a defined plane will not be drawn.'),
    'gui_num_workers': type_utils.GUIDocInfo('Number of workers','Each action is processed by a worker and each worker can handle one action at a time. Having more workers means more actions are processed simultaneously, but having too many will not provide any gain and might freeze the program and your whole computer. Since much of the processing utilizes more than one processor thread, set this value to significantly less than the number of threads available in your system. NB: If you currently have running or enqueued jobs, the number of workers will only be changed once all have completed or are cancelled.'),
}
if _missing_params:=[k for k in _params if k not in study_parameter_doc and k not in ['self','session_def','planes','individual_markers','working_directory','strict_check']]:
    raise NotImplementedError('Documentation missing for parameters:\n- '+'\n- '.join(_missing_params))
del _params
del _missing_params

def guess_config_dir(working_dir: str|pathlib.Path, config_dir_name: str = "config", json_file_name: str = Study.default_json_file_name) -> pathlib.Path:
    # can be invoked with either:
    # 1. the project folder;
    # 2. a session's working directory; or
    # 3. a recording's directory in a session's working directory.
    # So try three levels
    for i in range(3):
        if i>0:
            # try again in parent directory
            working_dir = working_dir.parent
        test_dir = working_dir / config_dir_name
        if not test_dir.is_dir():
            continue
        test_file = test_dir / json_file_name
        if test_file.is_file():
            return test_dir

    raise RuntimeError('config directory not found')


class OverrideLevel(enum.Enum):
    Session     = enum.auto()
    Recording   = enum.auto()
    FunctionArgs= enum.auto()

class StudyOverride:
    default_json_file_name = 'study_def_override.json'

    @staticmethod
    def get_allowed_parameters(level: OverrideLevel, recording_type: session.RecordingType|None = None) -> tuple[list[str],set[str]]:
        # NB: list instead of set as want to keep ordering
        all_params = list(study_parameter_types.keys())
        exclude = {'self', 'session_def', 'planes', 'individual_markers', 'working_directory', 'import_known_custom_eye_trackers'}
        # above is Session-level disallowed parameters. Depending on level, disallow more
        if level in [OverrideLevel.Recording, OverrideLevel.FunctionArgs]:
            # these make no sense on a recording level as they are settings for
            # processing functions that run on a whole session at once. As function
            # arguments they may make sense depending on the processing function that
            # is being called, but we cannot differentiate, so reject to be conservative
            # use whitelist
            if recording_type==session.RecordingType.Camera:
                # if for a camera recording, almost no parameters make sense to set
                include = {'auto_code_sync_points', 'auto_code_episodes'}
            else:
                include = {'get_cam_movement_for_et_sync_method','get_cam_movement_for_et_sync_function',
                           'auto_code_sync_points', 'auto_code_episodes',
                           'validate_do_global_shift', 'validate_max_dist_fac', 'validate_dq_types', 'validate_allow_dq_fallback', 'validate_include_data_loss', 'validate_I2MC_settings', 'validate_dynamic_skip_first_duration', 'validate_dynamic_max_gap_duration', 'validate_dynamic_min_duration'}
            exclude = set(all_params)-include
        allowed_params = [a for a in all_params if a not in exclude]
        return allowed_params, exclude

    def __init__(self, level: OverrideLevel, recording_type: session.RecordingType|None = None, **kwargs):
        self.override_level = level
        self.recording_type = recording_type
        self._allowed_params, self._excluded_parameters = self.get_allowed_parameters(level, recording_type)
        self._overridden_params: list[str] = []
        for p in self._allowed_params:
            self.clear_override(p)
        def typecheck_exception_handler(exc: typeguard.TypeCheckError, key: str):
            e = typeguard.TypeCheckError(*exc.args)
            e.append_path_element(f'argument "{key}" {self._get_err_msg()} ({exc._path[0]})')
            raise e from None
        kwargs = StudyOverride._fix_typing(kwargs)
        for p in kwargs:
            self._check_parameter(p, f"{StudyOverride.__name__}.__init__(): ")
            # special case: for dict-like object we can unset specific fields, so allow those by skipping check for them
            check_val = kwargs[p]
            if isinstance(check_val,dict) or typing.is_typeddict(check_val) or typed_dict_defaults.is_typeddictdefault(check_val) or type_utils.is_NamedTuple_type(check_val):
                check_val = {k:check_val[k] for k in check_val if check_val[k] is not None}
            typeguard.check_type(check_val, study_parameter_types[p], typecheck_fail_callback=lambda x,_: typecheck_exception_handler(x,p), collection_check_strategy=typeguard.CollectionCheckStrategy.ALL_ITEMS)
            setattr(self,p,kwargs[p])

    def __setattr__(self, name, value):
        if name.startswith('_') or name in {'override_level', 'recording_type'}:
            super(StudyOverride, self).__setattr__(name, value)
            return

        self._check_parameter(name)
        super(StudyOverride, self).__setattr__(name, value)
        if name not in self._overridden_params:
            self._overridden_params.append(name)

    def clear_override(self, name):
        self._check_parameter(name)
        setattr(self,name,None)
        if name in self._overridden_params:
            self._overridden_params.remove(name)

    def _check_parameter(self, name: str, error_prefix=''):
        if name in self._excluded_parameters:
            raise ValueError(f"{error_prefix}You are not allowed to override the '{name}' parameter of a {Study.__name__} class {self._get_err_msg()}")
        if name not in self._allowed_params:
            raise ValueError(f"{error_prefix}Got an unknown parameter '{name}'")

    def _get_err_msg(self):
        if self.override_level==OverrideLevel.FunctionArgs:
            err_text = 'in the parameter overrides provided as extra arguments to the processing function'
        else:
            err_text = f'in the {self.override_level.name}-level parameter overrides'
        if self.recording_type is not None:
            err_text += f' for a {self.recording_type.value} recording'
        return err_text

    def apply(self, study: Study, strict_check=True) -> Study:
        study = copy.deepcopy(study)
        study = _apply_impl(study, {p: getattr(self,p) for p in self._overridden_params}, study_parameter_types)
        # check resulting study is valid
        try:
            study.check_valid(strict_check)
        except Exception as oe:
            raise ValueError(f'Study setup became invalid {self._get_err_msg()}: {str(oe)}').with_traceback(oe.__traceback__) from None
        return study

    def store_as_json(self, path: str | pathlib.Path):
        path = pathlib.Path(path)
        if path.is_dir():
            path = path / self.default_json_file_name
        to_dump = {p:getattr(self,p) for p in self._overridden_params}
        if 'planes_per_episode' in to_dump:
            to_dump['planes_per_episode'] = [(k, to_dump['planes_per_episode'][k]) for k in to_dump['planes_per_episode']]   # pack as list of tuples for storage
        json.dump(to_dump, path)

    @staticmethod
    def load_from_json(level: OverrideLevel, path: str | pathlib.Path, recording_type: session.RecordingType|None = None) -> 'StudyOverride':
        path = pathlib.Path(path)
        if path.is_dir():
            path = path / StudyOverride.default_json_file_name
        kwds = json.load(path)
        if 'planes_per_episode' in kwds:
            # stored as list of tuples, unpack
            kwds['planes_per_episode'] = {k:v for k,v in kwds['planes_per_episode']}
        # help with enum roundtrip
        if 'episodes_to_code' in kwds:
            kwds['episodes_to_code'] = {annotation.Event(e) for e in kwds['episodes_to_code']}
        if 'validate_dq_types' in kwds:
            kwds['validate_dq_types']= {DataQualityType(d) for d in kwds['validate_dq_types']}
        if 'mapped_video_which_gaze_type_on_plane' in kwds:
            kwds['mapped_video_which_gaze_type_on_plane'] = gaze_worldref.Type(kwds['mapped_video_which_gaze_type_on_plane'])
        return StudyOverride(level, recording_type, **kwds)

    @staticmethod
    def from_study_diff(config: Study, parent_config: Study, level: OverrideLevel, recording_type: session.RecordingType|None = None) -> 'StudyOverride':
        fields = StudyOverride.get_allowed_parameters(level, recording_type)[0]
        kwds = _study_diff_impl(config, parent_config, fields)
        return StudyOverride(level, recording_type, **kwds)

    @staticmethod
    def _fix_typing(kwds: dict[str,Any]) -> dict[str,Any]:
        if 'get_cam_movement_for_et_sync_function' in kwds and kwds['get_cam_movement_for_et_sync_function'] is not None:
            kwds['get_cam_movement_for_et_sync_function'] = CamMovementForEtSyncFunction(**kwds['get_cam_movement_for_et_sync_function'])
        if 'auto_code_sync_points' in kwds and kwds['auto_code_sync_points'] is not None:
            kwds['auto_code_sync_points'] = AutoCodeSyncPoints(**kwds['auto_code_sync_points'])
        if 'auto_code_episodes' in kwds and kwds['auto_code_episodes'] is not None:
            kwds['auto_code_episodes'] = {e: AutoCodeEpisodes(**kwds['auto_code_episodes'][e]) for e in kwds['auto_code_episodes']}
        if 'validate_I2MC_settings' in kwds and kwds['validate_I2MC_settings'] is not None:
            kwds['validate_I2MC_settings'] = I2MCSettings(**kwds['validate_I2MC_settings'])
        if 'mapped_video_recording_colors' in kwds and kwds['mapped_video_recording_colors'] is not None:
            kwds['mapped_video_recording_colors'] = {k: None if kwds['mapped_video_recording_colors'][k] is None else RgbColor(*kwds['mapped_video_recording_colors'][k]) for k in kwds['mapped_video_recording_colors']}
        for k in ['mapped_video_projected_vidPos_ray_color','mapped_video_projected_world_pos_color','mapped_video_projected_left_ray_color','mapped_video_projected_right_ray_color','mapped_video_projected_average_ray_color']:
            if k in kwds and kwds[k] is not None:
                kwds[k] = RgbColor(**kwds[k])
        return kwds

def _apply_impl(obj, overrides: dict[str,Any], annotations: dict[str,typing.Type]|None):
    for p in overrides:
        ori_val = obj.get(p,None) if isinstance(obj,dict) else getattr(obj,p)
        val = overrides[p]
        not_found= isinstance(obj,dict) and p not in obj
        just_set = ori_val is None and not not_found
        if not just_set and (isinstance(val,dict) or type_utils.is_NamedTuple_type(type(val))):
            # dict-like object: recurse
            val = _apply_impl(ori_val, {p2: val[p2] if isinstance(val,dict) else getattr(val,p2) for p2 in type_utils.get_fields(val) if (isinstance(val,dict) and p2 in val) or hasattr(val,p2)}, type_utils.get_annotations(ori_val))

        # special case: for dict-like object we can unset specific fields, so allow those by skipping check for them
        if not just_set and val is None and (annotations is None or p not in annotations or not gt_utils.unpack_none_union(annotations[p])[1]):
            if isinstance(obj,dict):
                if p in obj:
                    del obj[p]
            else:
                delattr(obj,p)
        else:
            if isinstance(obj,dict):
                # overwrite existing and add new dict keys
                obj[p] = val
            elif type_utils.is_NamedTuple_type(type(obj)):
                # named tuples are immutable, have to return new instance
                obj = obj._replace(**{p:val})
            else:
                setattr(obj,p,val)
    return obj

def _study_diff_impl(config: Study, parent_config: Study, fields: list[str]) -> dict[str,Any]:
    kwds: dict[str,Any] = {}
    for f in fields:
        val        =        config.get(f) if isinstance(       config,dict) else getattr(       config,f)
        parent_val = parent_config.get(f) if isinstance(parent_config,dict) else getattr(parent_config,f)
        if val!=parent_val:
            if parent_val is not None and (isinstance(val,dict) or typing.is_typeddict(val) or typed_dict_defaults.is_typeddictdefault(val) or type_utils.is_NamedTuple_type(val)):
                # need to recurse into object
                val = _study_diff_impl(val, parent_val, list(set(type_utils.get_fields(val))|set(type_utils.get_fields(parent_val))))
            kwds[f] = val
    return kwds

def load_override_and_apply(study: Study, level: OverrideLevel, override_path: str|pathlib.Path, recording_type: session.RecordingType|None = None, strict_check=True) -> Study:
    override_path = pathlib.Path(override_path)
    if override_path.is_dir():
        override_path = override_path / StudyOverride.default_json_file_name
    if not override_path.is_file():
        return study

    study_override = StudyOverride.load_from_json(level, override_path, recording_type)
    return study_override.apply(study, strict_check)

def load_or_create_override(level: OverrideLevel, override_path: str|pathlib.Path, recording_type: session.RecordingType|None = None) -> StudyOverride:
    override_path = pathlib.Path(override_path)
    if override_path.is_dir():
        override_path = override_path / StudyOverride.default_json_file_name

    if not override_path.is_file():
        return StudyOverride(level, recording_type)
    else:
        return StudyOverride.load_from_json(level, override_path, recording_type)

def apply_kwarg_overrides(study: Study, strict_check=True, **kwargs) -> Study:
    if not kwargs:
        return study
    overrides = StudyOverride(OverrideLevel.FunctionArgs, **kwargs)
    return overrides.apply(study, strict_check)

def read_study_config_with_overrides(config_path: str|pathlib.Path, overrides: dict[OverrideLevel, str|pathlib.Path]=None, recording_type: session.RecordingType|None = None, strict_check=True, **kwargs) -> Study:
    study = Study.load_from_json(config_path)
    if overrides:
        for l in [OverrideLevel.Session, OverrideLevel.Recording]:
            if l in overrides:
                study = load_override_and_apply(study, l, overrides[l], recording_type, strict_check)
    if kwargs:
        study = apply_kwarg_overrides(study, strict_check, **kwargs)
    return study



from glassesTools import aruco
from collections import defaultdict

def format_duplicate_markers_msg(markers: set[tuple[int,int]]):
    # NB: input should be dictionary families, not dicts themselves
    # organize per dictionary family
    dict_markers: dict[int,list[int]] = defaultdict(list)
    for m,d in markers:
        dict_markers[d].append(m)
    dict_markers = {d:sorted(dict_markers[d]) for d in dict_markers}
    out = ''
    for i,d in enumerate(dict_markers):
        s = 's' if len(dict_markers[d])>1 else ''
        ids = ', '.join((str(x) for x in dict_markers[d]))
        d_str,is_family = aruco.family_to_str[d]
        f_str = ' family' if is_family else ''
        msg = f'marker{s} {ids} for the {d_str} dictionary{f_str}'
        if i==0:
            out = msg
        elif i==len(dict_markers)-1:
            out += f' and {msg}'
        else:
            out += f', {msg}'
    return out

def format_marker_sequence_msg(marker_set: list[tuple[int,int]]):
    # NB: input should be dictionary families, not dicts themselves
    # turn each dict into a string/family
    marker_set_str: list[tuple[str,bool,int]] = []
    all_same_family_or_dict = len(set((x[0] for x in marker_set)))==1
    marker_set.sort(key=lambda x: x[0])
    if not all_same_family_or_dict:
        marker_set.sort(key=lambda x: x[1])
    for m in marker_set:
        d_str,is_family = aruco.family_to_str[m[1]]
        marker_set_str.append((d_str, is_family, m[0]))
    if all_same_family_or_dict:
        m_str = ', '.join((str(m[2]) for m in marker_set_str))
        m_str += ' from the ' + (f'{marker_set_str[0][0]} family' if marker_set_str[0][1] else f'{marker_set_str[0][0]} dict')
    else:
        m_str = ', '.join((f'{m[2]} ({m[0] + (" family" if m[1] else "")})' for m in marker_set_str))
    return m_str


import pathlib
import shutil
import numpy as np
import pandas as pd
import polars as pl
from collections import defaultdict

from glassesTools import annotation, gaze_worldref, marker as gt_marker, naming as gt_naming, process_pool

from .. import config, episode, naming, process, session


def run(working_dir: str|pathlib.Path, export_path: str|pathlib.Path, to_export: list[str], config_dir: str|pathlib.Path = None, **study_settings):
    working_dir = pathlib.Path(working_dir) # working directory of a session, not of a recording
    export_path = pathlib.Path(export_path)
    if config_dir is None:
        config_dir = config.guess_config_dir(working_dir)
    config_dir  = pathlib.Path(config_dir)
    print(f'processing: {working_dir.name}')

    # get settings for the study
    study_config = config.read_study_config_with_overrides(config_dir, {config.OverrideLevel.Session: working_dir}, **study_settings)
    # get session info
    session_info = session.Session.from_definition(study_config.session_def, working_dir)
    all_recs    = [r for r in session_info.recordings]
    et_recs     = [r for r in session_info.recordings if session_info.recordings[r].definition.type==session.RecordingType.Eye_Tracker]

    if 'plane_gaze' in to_export:
        export_plane_gaze(export_path, working_dir, study_config, et_recs)

    if 'gaze_overlay_video' in to_export:
        export_gazeOverlay_video(export_path, working_dir, et_recs)

    if 'mapped_gaze_video' in to_export:
        export_mappedGaze_video(export_path, working_dir, all_recs)

    # update state
    session.update_action_states(working_dir, process.Action.EXPORT_TRIALS, process_pool.State.Completed, study_config)


def export_plane_gaze(export_path: pathlib.Path, working_dir: pathlib.Path, study_config: config.Study, recs: list[str]):
    if annotation.Event.Trial not in study_config.planes_per_episode:
        raise ValueError('No planes are specified for mapping gaze to during trials, no gaze-on-plane data to export')

    planes = list(study_config.planes_per_episode[annotation.Event.Trial])

    # per recording, read the relevant files and put them all together
    for r in recs:
        # check if files needed for export are present, else skip
        if not all(((working_dir / r / f'{naming.world_gaze_prefix}{p}.tsv').is_file() for p in planes)):
            print(f'Warning: not all plane gaze files found for recording {r} in session {working_dir.name}. Skipping...')
            continue
        if not all((gt_marker.get_file_name(m.id, m.aruco_dict_id, working_dir/r).is_file() for m in study_config.individual_markers)):
            print(f'Warning: not all individual marker detection files found for recording {r} in session {working_dir.name}. Skipping...')
            continue
        # get trial coding
        # trial episodes are gotten from the reference recording if there is one and this is not the reference recording
        if study_config.sync_ref_recording and r!=study_config.sync_ref_recording:
            episodes = episode.list_to_marker_dict(episode.read_list_from_file(working_dir / study_config.sync_ref_recording / naming.coding_file), study_config.episodes_to_code)
            subset_var = 'frame_idx_ref'
        else:
            episodes = episode.list_to_marker_dict(episode.read_list_from_file(working_dir / r / naming.coding_file), study_config.episodes_to_code)
            subset_var = 'frame_idx'
        if annotation.Event.Trial not in episodes or not episodes[annotation.Event.Trial]:
            print(f'Warning: no {annotation.tooltip_map[annotation.Event.Trial]}s found in the coding file for recording {r} in session {working_dir.name}. Skipping...')
            continue
        episodes = episodes[annotation.Event.Trial]

        # get all gaze data
        plane_gazes = {p:pd.read_csv(working_dir / r / f'{naming.world_gaze_prefix}{p}.tsv',sep='\t', dtype=defaultdict(lambda: float, **gaze_worldref.Gaze._non_float)) for p in planes}

        # throw away unwanted columns
        if not study_config.export_output3D:
            for p in plane_gazes:
                # throw away all columns starting with gazePosCam or gazeOriCam
                plane_gazes[p] = plane_gazes[p].drop(columns=[c for c in plane_gazes[p].columns if c.startswith('gazePosCam') or c.startswith('gazeOriCam')])
        if not study_config.export_output2D:
            for p in plane_gazes:
                # throw away all columns starting with gazePosPlane2D
                plane_gazes[p] = plane_gazes[p].drop(columns=[c for c in plane_gazes[p].columns if c.startswith('gazePosPlane2D')])

        # rename putting plane name in there so that names are unique
        for p in plane_gazes:
            plane_gazes[p] = plane_gazes[p].rename(columns={c:f'{c[:7]}_{p}_{c[7:]}' for c in plane_gazes[p].columns if c.startswith('gazePos') or c.startswith('gazeOri')})

        # now merge
        for i in range(1,len(planes)):
            # NB: by not providing on, merge is done on intersection of columns (so that is all timestamp and frame_idx columns, which is what we want)
            plane_gazes[planes[0]] = plane_gazes[planes[0]].merge(plane_gazes[planes[i]], how='outer')
        plane_gazes = plane_gazes[planes[0]]

        # if there are individual markers, add them
        # load
        markers = {m.id: gt_marker.read_dataframe_from_file(m.id, m.aruco_dict_id, working_dir/r) for m in study_config.individual_markers}
        # recode to presence/absence if wanted
        if study_config.export_only_code_marker_presence:
            markers = gt_marker.code_for_presence(markers, allow_failed=True)
        else:
            # rename columns to unique names
            for i in markers:
                markers[i] = markers[i].rename(columns={c:f'marker_{i}_{c}' for c in markers[i].columns if c not in ['frame_idx']})

        # merge
        for i in markers:
            # NB: by not providing on, merge is done on intersection of columns (so that is all timestamp and frame_idx columns, which is what we want)
            plane_gazes = plane_gazes.merge(markers[i], how='left', on='frame_idx')
        # correct missing values in presence column to false
        if study_config.export_only_code_marker_presence:
            for i in markers:
                plane_gazes[f'marker_{i}_presence'] = plane_gazes[f'marker_{i}_presence'].notnull()

        # add scene and reference camera timestamp info, if present
        ts = pd.read_csv(working_dir / r / gt_naming.frame_timestamps_fname,sep='\t').rename(columns={'timestamp':'frame_ts'})
        plane_gazes = plane_gazes.merge(ts, how="left", on='frame_idx')
        to_move = 1
        if 'frame_idx_VOR' in plane_gazes.columns:
            ts = ts.rename(columns={'frame_idx':'frame_idx_VOR','frame_ts':'frame_ts_VOR'})
            plane_gazes = plane_gazes.merge(ts, how="left", on='frame_idx_VOR')
            to_move += 1
        if 'frame_idx_ref' in plane_gazes.columns:
            ts = pd.read_csv(working_dir / study_config.sync_ref_recording / gt_naming.frame_timestamps_fname,sep='\t').rename(columns={'frame_idx':'frame_idx_ref','timestamp':'frame_ts_ref','timestamp_stretched':'frame_ts_ref_stretched'})
            plane_gazes = plane_gazes.merge(ts, how="left", on='frame_idx_ref')
            to_move += len([c for c in ts.columns if c.startswith('frame_ts_')])
        # reorder to get ts columns in the right place
        cols= plane_gazes.columns.to_list()
        idx = max([cols.index(c) for c in cols if c.startswith('frame_idx')])+1
        cols= cols[:idx] + cols[-to_move:] + cols[idx:-to_move]
        plane_gazes = plane_gazes[cols]

        # add trial numbers
        idx = max([cols.index(c) for c in cols if c.startswith('frame_ts')])+1
        plane_gazes.insert(idx,'trial',np.int32(-1))
        for i,e in enumerate(episodes):
            sel = (plane_gazes[subset_var] >= e[0]) & (plane_gazes[subset_var] <= e[1])
            plane_gazes.loc[sel,'trial'] = i+1

        # store
        # write into df (use polars as that library saves to file waaay faster)
        plane_gazes = pl.from_pandas(plane_gazes)
        plane_gazes.write_csv(export_path / f'{naming.gaze_export_name}_{working_dir.name}_{r}.tsv', separator='\t', null_value='nan', float_precision=8)

def export_gazeOverlay_video(export_path: pathlib.Path, working_dir: pathlib.Path, recs: list[str]):
    for r in recs:
        inFile = working_dir/r/gt_naming.gaze_overlay_video_file
        if not inFile.is_file():
            continue
        shutil.copy2(inFile, export_path / f'gazeOverlay_{working_dir.name}_{r}.mp4')

def export_mappedGaze_video(export_path: pathlib.Path, working_dir: pathlib.Path, recs: list[str]):
    for r in recs:
        inFile = working_dir/r/naming.mapped_gaze_video
        if not inFile.is_file():
            continue
        shutil.copy2(inFile, export_path / f'mappedGaze_{working_dir.name}_{r}.mp4')


from enum import auto
import pathlib
import typeguard
import shutil

from glassesTools import camera_recording, importing, json, naming, process_pool, utils
from glassesTools.recording import Recording as EyeTrackerRecording
from glassesTools.camera_recording import Recording as CameraRecording

from . import process


class RecordingType(utils.AutoName):
    Eye_Tracker = auto()
    Camera      = auto()
json.register_type(json.TypeEntry(RecordingType,'__enum.session.RecordingType__', utils.enum_val_2_str, lambda x: getattr(RecordingType, x.split('.')[1])))
recording_types = [r for r in RecordingType]


class RecordingDefinition:
    cal_file_name = naming.scene_camera_calibration_fname
    @typeguard.typechecked
    def __init__(self, name:str, type:RecordingType):
        self.name = name
        self.type = type

    def set_default_cal_file(self, cal_path: str|pathlib.Path, rec_def_path: str|pathlib.Path):
        cal_path = pathlib.Path(cal_path)
        rec_def_path = pathlib.Path(rec_def_path)
        shutil.copyfile(str(cal_path), str(rec_def_path / f'{self.name}_{RecordingDefinition.cal_file_name}'))

    def get_default_cal_file(self, rec_def_path: str|pathlib.Path) -> pathlib.Path|None:
        cal_path = pathlib.Path(rec_def_path) / f'{self.name}_{RecordingDefinition.cal_file_name}'
        if cal_path.is_file():
            return cal_path
        return None

    def remove_default_cal_file(self, rec_def_path: str|pathlib.Path):
        cal_path = pathlib.Path(rec_def_path) / f'{self.name}_{RecordingDefinition.cal_file_name}'
        cal_path.unlink(missing_ok=True)

json.register_type(json.TypeEntry(RecordingDefinition,'__session.RecordingDefinition__',lambda x: {'name': x.name, 'type': x.type}, lambda x: RecordingDefinition(name=x['name'], type=RecordingType(x['type']))))


class Recording:
    status_file_name = 'recording.gazeMapper'

    @typeguard.typechecked
    def __init__(self, definition: RecordingDefinition, info:EyeTrackerRecording|CameraRecording|None=None):
        self.definition = definition
        self.info       = info

        self.state: dict[process.Action, process_pool.State] = {}
        if not self.info.working_directory:
            # don't create action states file if this recording only exists in memory (still needs to be imported)
            self.state = _get_not_run_action_states(True)
        else:
            self.load_action_states(True, True)

    def load_action_states(self, create_if_missing: bool, upgrade_if_needed: bool):
        self.state |= get_action_states(self.info.working_directory, for_recording=True, create_if_missing=create_if_missing, upgrade_if_needed=upgrade_if_needed)
json.register_type(json.TypeEntry(Recording,'__session.Recording__',lambda x: {'defition': x.defition, 'info': x.info}, lambda x: Recording(**x)))

def read_recording_info(working_dir: pathlib.Path, rec_type: RecordingType) -> tuple[EyeTrackerRecording|CameraRecording, pathlib.Path]:
    if rec_type==RecordingType.Camera:
        rec_info = CameraRecording.load_from_json(working_dir)
    elif rec_type==RecordingType.Eye_Tracker:
        rec_info = EyeTrackerRecording.load_from_json(working_dir)
    return rec_info, get_video_path(rec_info)

def get_video_path(rec_info: EyeTrackerRecording|CameraRecording) -> pathlib.Path:
    if isinstance(rec_info, CameraRecording):
        return rec_info.get_video_path()
    elif isinstance(rec_info, EyeTrackerRecording):
        return rec_info.get_scene_video_path()

class SessionDefinition:
    default_json_file_name = 'session_def.json'

    @typeguard.typechecked
    def __init__(self, recordings: list[RecordingDefinition]|None=None):
        if recordings is None:
            recordings = []
        self.recordings = recordings

    def add_recording_def(self, recording: RecordingDefinition):
        self.recordings.append(recording)

    def get_recording_def(self, which: str) -> RecordingDefinition:
        for r in self.recordings:
            if r.name==which:
                return r
        raise ValueError(f'recording "{which}" not found')

    def is_known_recording(self, which: str) -> bool:
        for r in self.recordings:
            if r.name==which:
                return True
        return False

    def store_as_json(self, path: str | pathlib.Path):
        path = pathlib.Path(path)
        if path.is_dir():
            path /= self.default_json_file_name
        to_dump = {k:getattr(self,k) for k in vars(self) if not k.startswith('_')}
        json.dump(to_dump, path)

    @staticmethod
    def load_from_json(path: str | pathlib.Path) -> 'Session':
        path = pathlib.Path(path)
        if path.is_dir():
            path /= SessionDefinition.default_json_file_name
        kwds = json.load(path)
        return SessionDefinition(**kwds)
json.register_type(json.TypeEntry(SessionDefinition,'__session.SessionDefinition__',lambda x: {'recordings': x.recordings}, lambda x: SessionDefinition(**x)))


class Session:
    status_file_name = 'session.gazeMapper'

    @typeguard.typechecked
    def __init__(self, definition: SessionDefinition, name: str, working_directory: str|pathlib.Path|None = None, recordings: dict[str,Recording]|None = None):
        self.definition = definition
        self.name = name
        self.working_directory: pathlib.Path = pathlib.Path(working_directory) if working_directory else None
        if not recordings:
            recordings = {}
        self.recordings = recordings

        self.state: dict[process.Action, process_pool.State] = {}
        self.load_action_states(True, True)

    def create_working_directory(self, parent_directory: str|pathlib.Path):
        self.working_directory = pathlib.Path(parent_directory) / self.name
        if not self.working_directory.is_dir():
            self.working_directory.mkdir()
        if not (self.working_directory/Session.status_file_name).is_file():
            _create_action_states_file(self.working_directory, False)

    def import_recording(self, which: str, cam_cal_file: str|pathlib.Path=None, generic_device_name: str=None, **kwargs):
        from . import config
        rec_def = self.definition.get_recording_def(which)
        config_dir = config.guess_config_dir(self.working_directory)
        study_config = config.read_study_config_with_overrides(config_dir, {config.OverrideLevel.Session: self.working_directory}, **kwargs)

        # do import
        rec_info = self.recordings[which].info
        rec_info.working_directory = self.working_directory / rec_def.name
        if cam_cal_file is None:
            # get default calibration file, if there is one
            cam_cal_file = self.recordings[which].definition.get_default_cal_file(config_dir)
        if rec_def.type==RecordingType.Eye_Tracker:
            rec_info = importing.do_import(rec_info=rec_info, copy_scene_video=study_config.import_do_copy_video, source_dir_as_relative_path=study_config.import_source_dir_as_relative_path, cam_cal_file=cam_cal_file, device_name=generic_device_name)
        else:
            rec_info = camera_recording.do_import(rec_info=rec_info, copy_video=study_config.import_do_copy_video, source_dir_as_relative_path=study_config.import_source_dir_as_relative_path, cam_cal_file=cam_cal_file)
        self.update_recording_info(which, rec_info)    # the import call may have updated the info (e.g. filled in recording length that wasn't known from metadata). Update what we hold in memory
        # denote import finished
        _create_action_states_file(rec_info.working_directory, True)
        update_action_states(rec_info.working_directory, process.Action.IMPORT, process_pool.State.Completed, study_config)
        self.recordings[which].load_action_states(False, False)

    def add_recording_and_import(self, which: str, rec_info: EyeTrackerRecording|CameraRecording, cam_cal_file: str|pathlib.Path=None, **kwargs) -> Recording:
        rec = self.add_recording_from_info(which, rec_info)
        self.import_recording(which, cam_cal_file, **kwargs)
        return rec

    def load_existing_recordings(self):
        # load recordings that are present
        for r in self.definition.recordings:
            if (self.working_directory / r.name).is_dir():
                self.add_existing_recording(r.name)

    def load_recording_info(self, which) -> EyeTrackerRecording|CameraRecording:
        r_fold = self.working_directory / which
        if not r_fold.is_dir():
            return

        rec_def = self.definition.get_recording_def(which)
        if rec_def.type==RecordingType.Eye_Tracker:
            return EyeTrackerRecording.load_from_json(r_fold)
        else:
            return CameraRecording.load_from_json(r_fold)

    def add_existing_recording(self, which: str) -> Recording:
        if which in self.recordings:
            return  # nothing to do
        r_fold = self.working_directory / which
        if not r_fold.is_dir():
            return

        # get info about recording
        rec_info = self.load_recording_info(which)

        # add recording
        self.add_recording_from_info(which, rec_info)
        return self.recordings[which]

    def check_recording_info(self, which: str, rec_info: EyeTrackerRecording|CameraRecording):
        rec_def = self.definition.get_recording_def(which)
        if rec_def.type==RecordingType.Eye_Tracker:
            if not isinstance(rec_info,EyeTrackerRecording):
                raise TypeError(f"The provided rec_info is not for an eye tracker recording, but {which} is an eye tracker recording")
        elif rec_def.type==RecordingType.Camera:
            if not isinstance(rec_info,CameraRecording):
                raise TypeError(f"The provided rec_info is not for a camera recording, but {which} is a camera recording")

    def update_recording_info(self, which: str, rec_info: EyeTrackerRecording|CameraRecording):
        if which not in self.recordings:
            return
        self.check_recording_info(which, rec_info)
        self.recordings[which].info = rec_info

    def add_recording_from_info(self, which: str, rec_info: EyeTrackerRecording|CameraRecording) -> Recording:
        rec_def = self.definition.get_recording_def(which)
        self.check_recording_info(which, rec_info)
        self.recordings[which] = Recording(rec_def, rec_info)
        return self.recordings[which]

    def num_present_recordings(self) -> int:
        return sum((r.name in self.recordings for r in self.definition.recordings))

    def has_all_recordings(self) -> bool:
        return all((r.name in self.recordings for r in self.definition.recordings))

    def missing_recordings(self, rec_type: RecordingType|None=None) -> list[str]:
        return [r.name for r in self.definition.recordings if r.name not in self.recordings and (rec_type is None or r.type==rec_type)]

    # state of processing actions on recordings in a session
    def load_action_states(self, create_if_missing: bool, upgrade_if_needed: bool):
        if self.working_directory.is_dir():
            self.state |= get_action_states(self.working_directory, for_recording=False, create_if_missing=create_if_missing, upgrade_if_needed=upgrade_if_needed)

    def is_action_completed(self, action: process.Action) -> bool:
        if process.is_session_level_action(action):
            return self.state[action]==process_pool.State.Completed
        else:
            return all((self.recordings[r].state[action]==process_pool.State.Completed for r in self.recordings))

    def action_completed_num_recordings(self, action: process.Action) -> int:
        if process.is_session_level_action(action):
            raise ValueError('The status of session-level actions cannot be listed per recording')

        return sum([self.recordings[r].state[action]==process_pool.State.Completed for r in self.recordings])

    @staticmethod
    def from_definition(definition: SessionDefinition|None, path: str | pathlib.Path) -> 'Session':
        # for loading a recording directory that doesn't contain a session json file
        # use the provided definition instead. If no session definition, try to load it
        if definition is None:
            from . import config
            config_dir = config.guess_config_dir(path)
            definition = config.Study.load_from_json(config_dir).session_def
        sess = Session(definition, name=path.name, working_directory=path)
        sess.load_existing_recordings()
        return sess


def get_session_from_directory(path: str|pathlib.Path, session_def: SessionDefinition|None=None) -> Session:
    # try to get config, we'll need that to load recording sessions
    if session_def is None:
        from . import config
        config_dir = config.guess_config_dir(path)
        session_def = config.Study.load_from_json(config_dir).session_def

    if not path.is_dir():
        raise RuntimeError('The provided path should be a directory')

    f = path / Session.status_file_name
    if not f.is_file():
        raise RuntimeError('The provided path is not a session, cannot load')

    return Session.from_definition(session_def, path)

def get_sessions_from_project_directory(path: str|pathlib.Path, session_def: SessionDefinition|None=None) -> list[Session]:
    path = pathlib.Path(path)

    # try to get config, we'll need that to load recording sessions
    if session_def is None:
        from . import config
        config_dir = config.guess_config_dir(path)
        session_def = config.Study.load_from_json(config_dir).session_def

    # iterate through all folders in the provided path and check if the folder contains
    # a session marker file. If so, try to to load the folder as a session, ignoring errors
    sessions: list[Session] = []
    for d in path.iterdir():
        try:
            sess = get_session_from_directory(d, session_def)
        except:
            pass
        else:
            sessions.append(sess)

    return sessions


def _get_action_status_fname(for_recording: bool) -> str:
    if for_recording:
        return Recording.status_file_name
    else:
        return Session.status_file_name

def _get_not_run_action_states(for_recording: bool) -> dict[process.Action, process_pool.State]:
    if for_recording:
        filt = lambda x: not process.is_session_level_action(x)
    else:
        filt = lambda x:     process.is_session_level_action(x)
    return {k:process_pool.State.Not_Run for k in process.Action if filt(k)}

def _create_action_states_file(file: pathlib.Path, for_recording: bool):
    if file.is_dir():
        file /= _get_action_status_fname(for_recording)
    action_states = _get_not_run_action_states(for_recording)
    _write_action_states_to_file(file, action_states)

def _write_action_states_to_file(file: pathlib.Path, action_states: dict[process.Action, process_pool.State]):
    action_states = {utils.enum_val_2_str(k):action_states[k] for k in action_states}    # turn key into string so it can be stored in a json file
    json.dump(action_states, file)

def _read_action_states(file: pathlib.Path) -> dict[process.Action, process_pool.State]:
    if not file.is_file():
        return None

    action_states = json.load(file)
    return {process.action_str_to_enum_val(k): process_pool.State(action_states[k]) for k in action_states}  # turn key from string back into enum instance, and same for state value

def _upgrade_action_states(file: pathlib.Path, action_states: dict[process.Action, process_pool.State], for_recording: bool) -> dict[process.Action, process_pool.State]:
    if file.is_dir():
        file /= _get_action_status_fname(for_recording)
    expected_action_states = _get_not_run_action_states(for_recording)
    # remove no longer available action_states
    upgraded_action_states = {a:action_states[a] for a in action_states if a in expected_action_states}
    # add missing action_states
    upgraded_action_states |= {a:expected_action_states[a] for a in expected_action_states if a not in upgraded_action_states}
    # store if anything has changed
    if set(upgraded_action_states.keys())!=set(action_states.keys()):
        _write_action_states_to_file(file, upgraded_action_states)
    return upgraded_action_states

def get_action_states(working_dir: str|pathlib.Path, for_recording: bool, create_if_missing = False, skip_if_missing=False, upgrade_if_needed=False) -> dict[process.Action, process_pool.State]:
    working_dir = pathlib.Path(working_dir)
    file = working_dir / _get_action_status_fname(for_recording)
    action_states = _read_action_states(file)

    if action_states is not None:
        if upgrade_if_needed:
            action_states = _upgrade_action_states(file, action_states, for_recording)
    else:
        if create_if_missing:
            _create_action_states_file(file, for_recording)
            return _read_action_states(file)
        elif not skip_if_missing:
            raise FileNotFoundError(f'Action states file {file} was not found')
    return action_states

def _apply_mutations_and_store(file, action_state_mutations, skip_if_missing=False):
    action_states = _read_action_states(file)
    if action_states is None and not skip_if_missing:
        raise FileNotFoundError(f'Action states file {file} was not found')

    # apply mutations
    for a in action_state_mutations:
        action_states[a] = action_state_mutations[a]

    _write_action_states_to_file(file, action_states)

def update_action_states(working_dir: str|pathlib.Path, action: process.Action, state: process_pool.State, study_config: 'config.Study', skip_if_missing=False, unchanged=False) -> dict[process.Action, process_pool.State]:
    for_recording = not process.is_session_level_action(action)

    if unchanged:
        # just update state of this task, don't cascade
        action_state_mutations = {action: state}
        if for_recording:
            file = working_dir / _get_action_status_fname(True)
            _apply_mutations_and_store(file, action_state_mutations, skip_if_missing=skip_if_missing)
        else:
            file = working_dir / _get_action_status_fname(False)
            _apply_mutations_and_store(file, action_state_mutations, skip_if_missing=skip_if_missing)
        return

    # determine state mutations
    action_state_mutations, for_all_recs = process.action_update_and_invalidate(action, state, study_config)
    # split in session-level and recording-level actions, report them separately
    session_state_mutations   = {a:action_state_mutations[a] for a in action_state_mutations if     process.is_session_level_action(a)}
    recording_state_mutations = {a:action_state_mutations[a] for a in action_state_mutations if not process.is_session_level_action(a)}

    # get which recordings to apply to
    session_dir = working_dir.parent if for_recording else working_dir
    if not for_recording or for_all_recs:
        sess = get_session_from_directory(session_dir)
        recs = list(sess.recordings.keys())
    else:
        recs = [working_dir.name]
    # and apply and store recording-level mutations
    for r in recs:
        f = session_dir / r / _get_action_status_fname(True)
        _apply_mutations_and_store(f, recording_state_mutations, skip_if_missing=skip_if_missing)
    # also apply and store session-level mutations
    if session_state_mutations:
        f = session_dir /     _get_action_status_fname(False)
        _apply_mutations_and_store(f,   session_state_mutations, skip_if_missing=skip_if_missing)

    return session_state_mutations, recording_state_mutations


import threading
from imgui_bundle import imgui, icons_fontawesome_6 as ifa6
import typing

import glassesTools

from ... import process, session


class ColumnSpec(typing.NamedTuple):
    name: str
    flags: int
    display_func:  typing.Callable[[session.Session],None]
    sort_key_func: typing.Callable[[int],typing.Any]
    header_lbl: str|None=None   # if set, different string than name is used for the column header. Works only for non-angled headers

class List:
    def __init__(self,
                 items: dict[int|str, session.Session],
            items_lock: threading.Lock,
        selected_items: dict[int|str, bool],
        info_callback: typing.Callable = None,
        draw_action_status_callback: typing.Callable = None,
        item_context_callback: typing.Callable = None):

        self.items = items
        self.selected_items = selected_items
        self.items_lock     = items_lock

        self.info_callback  = info_callback
        self.draw_action_status_callback = draw_action_status_callback
        self.item_context_callback  = item_context_callback

        self.sorted_ids: list[int|str] = []
        self._last_clicked_id: int|str = None
        self._require_sort: bool = True

        self.display_actions: list[process.Action] = []
        self._view_column_count_base = 3    # selector, name, recordings
        self._view_column_count = None

        self._extra_columns: list[ColumnSpec] = None

        self._last_y = None

        self.table_flags: int = (
            imgui.TableFlags_.scroll_x |
            imgui.TableFlags_.scroll_y |
            imgui.TableFlags_.hideable |
            imgui.TableFlags_.sortable |
            imgui.TableFlags_.sort_multi |
            imgui.TableFlags_.reorderable |
            imgui.TableFlags_.sizing_fixed_fit |
            imgui.TableFlags_.highlight_hovered_column #|
            #imgui.TableFlags_.borders_inner_v |
            #imgui.TableFlags_.no_borders_in_body_until_resize
        )

    def set_actions_to_show(self, actions: set[process.Action]):
        self.display_actions = [k for k in process.Action if k in actions] # filter out crap and fix display order
        self._determine_column_count()

    def set_extra_columns(self, columns: list[ColumnSpec]|None):
        self._extra_columns = columns
        self._determine_column_count()

    def _determine_column_count(self):
        self._view_column_count = self._view_column_count_base + len(self.display_actions) + (0 if self._extra_columns is None else len(self._extra_columns))

    def draw(self, limit_outer_size=False):
        with self.items_lock:
            num_items = len(self.items)
        if num_items==0:
            imgui.text_wrapped('There are no sessions')
            return
        outer_size = imgui.ImVec2(0,0)
        if limit_outer_size and self._last_y is not None:
            outer_size.y = self._last_y+imgui.get_style().item_spacing.y
        if imgui.begin_table(
            f"##session_list",
            columns=self._view_column_count,
            flags=self.table_flags,
            outer_size=outer_size
        ):
            frame_height = imgui.get_frame_height()

            # Setup
            checkbox_width = frame_height-2*imgui.get_style().frame_padding.y
            imgui.table_setup_column("Selector", imgui.TableColumnFlags_.no_hide | imgui.TableColumnFlags_.no_sort | imgui.TableColumnFlags_.no_reorder | imgui.TableColumnFlags_.no_header_label, init_width_or_weight=checkbox_width)  # 0
            imgui.table_setup_column("Name", imgui.TableColumnFlags_.default_sort | imgui.TableColumnFlags_.no_hide)  # 1
            imgui.table_setup_column("Recordings", imgui.TableColumnFlags_.angled_header)  # 2
            # 3+
            for k in self.display_actions:
                imgui.table_setup_column(k.displayable_name, imgui.TableColumnFlags_.angled_header)
            if self._extra_columns is not None:
                for c in self._extra_columns:
                    imgui.table_setup_column(c.name, c.flags)

            # Sticky column headers and selector row
            if imgui.table_get_column_flags(0) & imgui.TableColumnFlags_.is_enabled:
                imgui.table_setup_scroll_freeze(1, 2)  # Sticky column headers and selector row
            else:
                imgui.table_setup_scroll_freeze(0, 2)  # Sticky column headers

            # Sorting
            with self.items_lock:
                sort_specs = imgui.table_get_sort_specs()
                if set(self.sorted_ids) != set(self.selected_items.keys()):
                    self._require_sort = True
                self._sort_items(sort_specs)

                # Headers
                imgui.table_angled_headers_row()
                imgui.table_next_row(imgui.TableRowFlags_.headers)
                for c_idx in range(self._view_column_count):
                    if not imgui.table_set_column_index(c_idx):
                        continue
                    if c_idx==0:  # checkbox column: reflects whether all, some or none of visible sessions are selected, and allows selecting all or none
                        # get state
                        num_selected = sum([self.selected_items[iid] for iid in self.sorted_ids])
                        if num_selected==0:
                            # none selected
                            multi_selected_state = -1
                        elif num_selected==len(self.sorted_ids):
                            # all selected
                            multi_selected_state = 1
                        else:
                            # some selected
                            multi_selected_state = 0

                        if multi_selected_state==0:
                            imgui.push_item_flag(imgui.internal.ItemFlagsPrivate_.mixed_value, True)
                        clicked, new_state = glassesTools.gui.utils.my_checkbox(f"##header_checkbox", multi_selected_state==1, frame_size=(0,0), frame_padding_override=(imgui.get_style().frame_padding.x/2,0), do_vertical_align=False)
                        if multi_selected_state==0:
                            imgui.pop_item_flag()

                        if clicked:
                            glassesTools.utils.set_all(self.selected_items, new_state, subset = self.sorted_ids)
                    else:
                        # action status columns or extra columns
                        if c_idx<3:
                            column_name = ("Name","Recordings")[c_idx-1]
                        else:
                            idx = c_idx-self._view_column_count_base
                            if idx<len(self.display_actions):
                                # action columns
                                column_name = self.display_actions[idx].displayable_name
                            else:
                                # extra columns
                                idx = idx-len(self.display_actions)
                                column_name = self._extra_columns[idx].header_lbl if self._extra_columns[idx].header_lbl is not None else self._extra_columns[idx].name
                        if imgui.table_get_column_flags(c_idx) & imgui.TableColumnFlags_.no_header_label:
                            column_name = '##'+column_name
                        imgui.table_header(column_name)

                # Loop rows
                any_selectable_clicked = False
                if self.sorted_ids and self._last_clicked_id not in self.sorted_ids:
                    # default to topmost if last_clicked unknown, or no longer on screen due to filter
                    self._last_clicked_id = self.sorted_ids[0]
                for iid in self.sorted_ids:
                    imgui.table_next_row()

                    item = self.items[iid]
                    num_columns_drawn = 0
                    selectable_clicked = False
                    checkbox_clicked, checkbox_hovered = False, False
                    config_button_hovered = False
                    has_drawn_hitbox = False
                    for ri in range(self._view_column_count+1):
                        if not (imgui.table_get_column_flags(ri) & imgui.TableColumnFlags_.is_enabled):
                            continue
                        imgui.table_set_column_index(ri)

                        # Row hitbox
                        if not has_drawn_hitbox:
                            # hitbox needs to be drawn before anything else on the row so that, together with imgui.set_item_allow_overlap(), hovering button
                            # or checkbox on the row will still be correctly detected.
                            # this is super finicky, but works. The below together with using a height of frame_height+cell_padding_y
                            # makes the table row only cell_padding_y/2 longer. The whole row is highlighted correctly
                            cell_padding_y = imgui.get_style().cell_padding.y
                            cur_pos_y = imgui.get_cursor_pos_y()
                            imgui.set_cursor_pos_y(cur_pos_y - cell_padding_y/2)
                            imgui.push_style_var(imgui.StyleVar_.frame_border_size, 0.)
                            imgui.push_style_var(imgui.StyleVar_.frame_padding    , (0.,0.))
                            imgui.push_style_var(imgui.StyleVar_.item_spacing     , (0.,cell_padding_y))
                            selectable_clicked, selectable_out = imgui.selectable(f"##{iid}_hitbox", self.selected_items[iid], flags=imgui.SelectableFlags_.span_all_columns|imgui.SelectableFlags_.allow_overlap|imgui.internal.SelectableFlagsPrivate_.select_on_click, size=(0,frame_height+cell_padding_y))
                            imgui.set_cursor_pos_y(cur_pos_y)   # instead of imgui.same_line(), we just need this part of its effect
                            imgui.pop_style_var(3)
                            selectable_right_clicked, selectables_edited = glassesTools.gui.utils.handle_item_hitbox_events(iid, self.selected_items, context_menu=self.item_context_callback)
                            self._require_sort |= selectables_edited
                            has_drawn_hitbox = True

                        if num_columns_drawn==1:
                            # (Invisible) button because it aligns the following draw calls to center vertically
                            imgui.push_style_var(imgui.StyleVar_.frame_border_size, 0.)
                            imgui.push_style_var(imgui.StyleVar_.frame_padding    , (0.,imgui.get_style().frame_padding.y))
                            imgui.push_style_var(imgui.StyleVar_.item_spacing     , (0.,imgui.get_style().item_spacing.y))
                            imgui.push_style_color(imgui.Col_.button, (0.,0.,0.,0.))
                            imgui.button(f"##{item.name}_id", size=(imgui.FLT_MIN, 0))
                            imgui.pop_style_color()
                            imgui.pop_style_var(3)

                            imgui.same_line()

                        match ri:
                            case 0:
                                # Selector
                                checkbox_clicked, checkbox_out = glassesTools.gui.utils.my_checkbox(f"##{iid}_selected", self.selected_items[iid], frame_size=(0,0), frame_padding_override=(imgui.get_style().frame_padding.x/2,imgui.get_style().frame_padding.y))
                                checkbox_hovered = imgui.is_item_hovered()
                            case 1:
                                # Name
                                if self.info_callback:
                                    if imgui.button(f"{ifa6.ICON_FA_CIRCLE_INFO}##{iid}_info"):
                                        self._show_item_info(iid)
                                    glassesTools.gui.utils.draw_hover_text('More info about this session, and access session-specific configuration', '')
                                    config_button_hovered = imgui.is_item_hovered()
                                    imgui.same_line()
                                imgui.text(item.name)
                            case 2:
                                # Number of recordings
                                n_rec = len(item.definition.recordings)
                                missing_recs = item.missing_recordings()
                                clr = glassesTools.gui.colors.error if missing_recs else glassesTools.gui.colors.ok
                                imgui.text_colored(clr, f'{n_rec-len(missing_recs)}/{n_rec}{" "+ifa6.ICON_FA_TRIANGLE_EXCLAMATION if missing_recs else ""}')
                                if missing_recs:
                                    glassesTools.gui.utils.draw_hover_text('missing recordings:\n'+'\n'.join(missing_recs), '')
                            case _:
                                # action status columns or extra columns
                                idx = ri-self._view_column_count_base
                                if idx<len(self.display_actions):
                                    # action columns
                                    self.draw_action_status_callback(item, self.display_actions[idx])
                                else:
                                    # extra columns
                                    idx = idx-len(self.display_actions)
                                    self._extra_columns[idx].display_func(item)
                        num_columns_drawn+=1

                    # handle selection logic
                    # NB: the part of this logic that has to do with right-clicks is in handle_item_hitbox_events()
                    # NB: any_selectable_clicked is just for handling clicks not on any item
                    any_selectable_clicked = any_selectable_clicked or selectable_clicked or selectable_right_clicked

                    self._last_clicked_id = glassesTools.gui.utils.selectable_item_logic(
                        iid, self.selected_items, self._last_clicked_id, self.sorted_ids,
                        selectable_clicked, selectable_out, overlayed_hovered=checkbox_hovered or config_button_hovered,
                        overlayed_clicked=checkbox_clicked, new_overlayed_state=checkbox_out
                        )

                    # further deal with doubleclick on item
                    if selectable_clicked and not checkbox_hovered: # don't enter this branch if interaction is with checkbox on the table row
                        if not imgui.get_io().key_ctrl and not imgui.get_io().key_shift and imgui.is_mouse_double_clicked(imgui.MouseButton_.left):
                            self._show_item_info(iid)

                self._last_y = imgui.get_cursor_pos().y
                last_cursor_y = imgui.get_cursor_screen_pos().y
                imgui.end_table()

                # handle click in table area outside header+contents:
                # deselect all, and if right click, show popup
                # check mouse is below bottom of last drawn row so that clicking on the one pixel empty space between selectables
                # does not cause everything to unselect or popup to open
                if imgui.is_item_clicked(imgui.MouseButton_.left) and not any_selectable_clicked and imgui.get_io().mouse_pos.y>last_cursor_y:  # NB: table header is not signalled by is_item_clicked(), so this works correctly
                    glassesTools.utils.set_all(self.selected_items, False)

            # show menu when right-clicking the empty space
            # TODO

    def _show_item_info(self, iid):
        if self.info_callback:
            self.info_callback(self.items[iid])

    def _sort_items(self, sort_specs_in: imgui.TableSortSpecs):
        if sort_specs_in.specs_dirty or self._require_sort:
            ids = list(self.items)
            sort_specs = [sort_specs_in.get_specs(i) for i in range(sort_specs_in.specs_count)]
            for sort_spec in reversed(sort_specs):
                match sort_spec.column_index:
                    case 1:     # Name
                        key = lambda iid: self.items[iid].name
                    case 2:     # Number of recordings
                        key = lambda iid: self.items[iid].num_present_recordings()
                    case _:     # status indicators
                        idx = sort_spec.column_index-self._view_column_count_base
                        if idx<len(self.display_actions):
                            # action columns
                            action = self.display_actions[sort_spec.column_index-self._view_column_count_base]
                            if process.is_session_level_action(action):
                                key = lambda iid: 999 if not self.items[iid].has_all_recordings() else self.items[iid].state[action]
                            else:
                                key = lambda iid: 999 if not self.items[iid].has_all_recordings() else self.items[iid].action_completed_num_recordings(action)
                        else:
                            # extra columns
                            idx = idx-len(self.display_actions)
                            key = self._extra_columns[idx].sort_key_func
                ids.sort(key=key, reverse=sort_spec.get_sort_direction()==imgui.SortDirection.descending)
            self.sorted_ids = ids
            sort_specs_in.specs_dirty = False
            self._require_sort = False



plane_pose_prefix   = 'planePose_'
world_gaze_prefix   = 'planeGaze_'
coding_file         = 'coding.tsv'
target_sync_file    = 'et_sync_target_file.tsv'
VOR_sync_file       = 'VOR_sync.tsv'
validation_prefix   = 'validate_'
mapped_gaze_video   = 'mappedGaze.mp4'
gaze_export_name    = 'planeGaze'


import pathlib
import pandas as pd
from collections import defaultdict

from glassesTools import annotation


class Episode:
    def __init__(self, event:annotation.Event, start_frame:int, end_frame:int=None):
        self.event          = event
        self.start_frame    = start_frame

        if annotation.type_map[event]==annotation.Type.Interval:
            if end_frame is None:
                raise ValueError(f"end frame expected for an interval-type episode ({annotation.tooltip_map[event]}), but not provided")
        if annotation.type_map[event]==annotation.Type.Point:
            if end_frame is not None:
                raise ValueError(f"end frame provided but not expected for a point-type episode ({annotation.tooltip_map[event]})")
        self.end_frame      = end_frame


# for dealing with lists of events
def read_list_from_file(fileName: str|pathlib.Path) -> list[Episode]:
    df = pd.read_csv(str(fileName), delimiter='\t', index_col=False, dtype=defaultdict(lambda: int, event=str, end_frame=pd.Int64Dtype()))
    df['event']     = [annotation.Event(x) for x in df['event'].values]
    df['end_frame'] = df['end_frame'].astype('object')
    df.loc[pd.isnull(df['end_frame']),'end_frame'] = None   # set missing to None
    return [Episode(**kwargs) for kwargs in df.to_dict(orient='records')]

def write_list_to_file(episodes: list[Episode],
                       fileName: str|pathlib.Path):
    if not episodes:
        return

    records = [{k:getattr(p,k) for k in vars(p) if not k.startswith('_')} for p in episodes]
    df = pd.DataFrame.from_records(records)
    df['event'] = [x.value for x in df['event'].values]

    # keep only columns to be written out and order them correctly
    df = df[['event','start_frame','end_frame']]
    df.to_csv(str(fileName), index=False, sep='\t', na_rep='nan')


def get_empty_marker_dict(episodes: list[annotation.Event]=None) -> dict[annotation.Event,list[list[int]]]:
    if not episodes:
        return {e:[] for e in annotation.Event}
    else:
        return {e:[] for e in annotation.Event if e in episodes}    # ensure return always has the same order

def list_to_marker_dict(episodes: list[Episode], expected_types: list[annotation.Event]=None) -> dict[annotation.Event,list[list[int]]]:
    e_dict = get_empty_marker_dict(expected_types)
    for e in episodes:
        if e.event not in e_dict:
            raise ValueError(f'{annotation.tooltip_map[e.event]}(s) found, but not expected (e.g. should not be coded for this study according to the study setup)')
        if e.end_frame is not None:
            e_dict[e.event].append([e.start_frame, e.end_frame])
        else:
            e_dict[e.event].append([e.start_frame])
    return {e:e_dict[e] for e in annotation.Event if e in e_dict}   # ensure return always has the same order

def marker_dict_to_list(episodes: dict[annotation.Event,list[int]|list[list[int]]]) -> list[Episode]:
    e_list: list[Episode] = []
    for e in episodes:
        if not episodes[e]:
            continue
        if isinstance(episodes[e][0],list):
            e_list.extend([Episode(e, *v) for v in episodes[e]])
        else:
            if annotation.type_map[e]==annotation.Type.Interval:
                for m in range(0,len(episodes[e])-1,2): # read in batches of two, and run until -1 to make sure we don't pick up incomplete intervals
                    e_list.append(Episode(e, *episodes[e][m:m+2]))
            else:
                e_list.extend([Episode(e,m) for m in episodes[e]])

    return sorted(e_list, key=lambda x: x.start_frame)


def is_in_interval(episodes: dict[annotation.Event,list[int]]|list[Episode], idx: int) -> dict[annotation.Event, bool]:
    if isinstance(episodes,dict):
        episodes = marker_dict_to_list(episodes)

    e_dict: dict[annotation.Event,bool] = {e:False for e in annotation.Event}
    for e in episodes:
        if annotation.type_map[e.event]==annotation.Type.Interval:
            if idx>=e.start_frame and idx<=e.end_frame:
                e_dict[e.event] = True
        else:
            if idx==e.start_frame:
                e_dict[e.event] = True
    return e_dict


import typing
import inspect
import dataclasses

from glassesTools import aruco

from . import typed_dict_defaults

ProblemDict = dict[str,typing.Union[None,str,'ProblemDict']]
NestedDict = dict[str,typing.Union[None,'NestedDict']]

@dataclasses.dataclass
class GUIDocInfo:
    display_string: str
    doc_str:        str
    children:       dict[str,'GUIDocInfo'] = dataclasses.field(default_factory=lambda: {})

ArucoDictType = typing.Literal[tuple(aruco.dict_id_to_str.keys())]


def merge_problem_dicts(a: ProblemDict, b: ProblemDict):
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_problem_dicts(a[key], b[key])
            elif isinstance(a[key], dict) or isinstance(b[key], dict):
                if isinstance(a[key], dict):
                    if 'problem_with_this_key' in a[key]:
                        a[key]['problem_with_this_key'] = '\n'.join([a[key]['problem_with_this_key'], b[key]])
                    else:
                        a[key]['problem_with_this_key'] = b[key]
                else:
                    temp = a[key]
                    a[key] = b[key].copy()
                    if 'problem_with_this_key' in a[key]:
                        a[key]['problem_with_this_key'] = '\n'.join([a[key]['problem_with_this_key'], temp])
                    else:
                        a[key]['problem_with_this_key'] = temp
            elif a[key] is None:
                a[key] = b[key]
            elif b[key] is None:
                pass    # do nothing
            else:
                a[key] = '\n'.join([a[key], b[key]])
        else:
            a[key] = b[key]
    return a


def is_NamedTuple_type(x):
  return (inspect.isclass(x) and issubclass(x, tuple) and
          hasattr(x, '_asdict') and callable(x._asdict) and
          hasattr(x, '__annotations__') and
          getattr(x, '_fields', None) is not None)

def get_fields(obj) -> list[str]|None:
    if not isinstance(obj, typing.Type):
        tobj = type(obj)
    else:
        tobj = obj
    if typing.is_typeddict(tobj):
        return list(obj.__annotations__.keys())
    elif typed_dict_defaults.is_typeddictdefault(tobj):
        return list(obj.__annotations__.keys())
    elif is_NamedTuple_type(tobj):
        return list(obj._fields)
    elif isinstance(obj, dict):
        return list(obj.keys())
    return None

def get_annotations(obj) -> dict[str, typing.Type]|None:
    if not isinstance(obj, typing.Type):
        tobj = type(obj)
    else:
        tobj = obj
    if typing.is_typeddict(tobj):
        return obj.__annotations__.copy()
    elif typed_dict_defaults.is_typeddictdefault(tobj):
        return obj.__annotations__.copy()
    elif is_NamedTuple_type(tobj):
        return obj.__annotations__.copy()
    elif isinstance(obj, dict):
        return {k:type(obj[k]) for k in obj}
    return None


__version__ = "1.8.0"

__title__ = "gazeMapper"
__description__ = "Automated world-based analysis of wearable eye tracker data."
__url__ = "https://github.com/dcnieho/gazeMapper"
__uri__ = __url__
__doc__ = __description__ + " <" + __url__ + ">"

__author__ = "Diederick Niehorster"
__email__ = "diederick_c.niehorster@humlab.lu.se"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2024 " + __author__



import pathlib
import numpy as np
import cv2
import copy
from imgui_bundle import imgui

from ffpyplayer.player import MediaPlayer

import sys

isMacOS = sys.platform.startswith("darwin")
if isMacOS:
    import AppKit

from glassesTools import annotation, drawing, gaze_headref, gaze_worldref, naming as gt_naming, ocv, pose as gt_pose, process_pool, propagating_thread, timestamps
from glassesTools.gui.video_player import GUI


from .. import config, episode, naming, plane, process, session, synchronization

# This script shows a video player that is used to indicate the interval(s)
# during which the poster should be found in the video and in later
# steps data quality computed. So this interval/these intervals would for
# instance be the exact interval during which the subject performs the
# validation task.
# This script can be run directly on recordings converted to the common format,
# but output from the detectMarkers and gazeToPoster actions (if available)
# will also be shown.

_event_type_to_key_map = {
    annotation.Event.Validate       : imgui.Key.v,
    annotation.Event.Sync_Camera    : imgui.Key.c,
    annotation.Event.Sync_ET_Data   : imgui.Key.e,
    annotation.Event.Trial          : imgui.Key.t,
}

def run(working_dir: str|pathlib.Path, config_dir: str|pathlib.Path = None, **study_settings):
    # if show_poster, also draw poster with gaze overlaid on it (if available)
    working_dir = pathlib.Path(working_dir)
    if config_dir is None:
        config_dir = config.guess_config_dir(working_dir)
    config_dir  = pathlib.Path(config_dir)

    print(f'processing: {working_dir.parent.name}/{working_dir.name}')

    # We run processing in a separate thread (GUI needs to be on the main thread for OSX, see https://github.com/pthom/hello_imgui/issues/33)
    gui = GUI(use_thread = False)
    gui.add_window(f'{working_dir.parent.name}, {working_dir.name}')

    proc_thread = propagating_thread.PropagatingThread(target=do_the_work, args=(working_dir, config_dir, gui), kwargs=study_settings, cleanup_fun=gui.stop)
    proc_thread.start()
    gui.start()
    proc_thread.join()


def do_the_work(working_dir: pathlib.Path, config_dir: pathlib.Path, gui: GUI, **study_settings):
    # get settings for the study
    study_config = config.read_study_config_with_overrides(config_dir, {config.OverrideLevel.Session: working_dir.parent, config.OverrideLevel.Recording: working_dir}, **study_settings)

    # get info about recording
    rec_def  = study_config.session_def.get_recording_def(working_dir.name)
    in_video = session.read_recording_info(working_dir, rec_def.type)[1]
    if rec_def.type==session.RecordingType.Camera:
        has_gaze, has_plane_gaze, has_plane_pose = False, False, False
        # no episode.Event.Sync_ET_Data for camera recordings, remove
        if annotation.Event.Sync_ET_Data in study_config.episodes_to_code:
            study_config.episodes_to_code.remove(annotation.Event.Sync_ET_Data)
        # no episode.Event.Validate for camera recordings, remove
        if annotation.Event.Validate in study_config.episodes_to_code:
            study_config.episodes_to_code.remove(annotation.Event.Validate)
    elif rec_def.type==session.RecordingType.Eye_Tracker:
        # Read gaze data
        has_gaze = True
        gazes = gaze_headref.read_dict_from_file(working_dir / gt_naming.gaze_data_fname, ts_column_suffixes=['VOR',''])[0]

        planes: set[str] = set()
        for e in [annotation.Event.Validate, annotation.Event.Trial]:
            if e in study_config.planes_per_episode:
                planes.update(study_config.planes_per_episode[e])
        if study_config.get_cam_movement_for_et_sync_method=='plane' and annotation.Event.Sync_ET_Data in study_config.planes_per_episode:
            planes.update(study_config.planes_per_episode[annotation.Event.Sync_ET_Data])

        # Read gaze on poster data, if available
        plane_files = [working_dir/f'{naming.world_gaze_prefix}{p}.tsv' for p in planes]
        plane_gazes = {p:gaze_worldref.read_dict_from_file(f) for p,f in zip(planes,plane_files) if f.is_file()}
        has_plane_gaze = not not plane_gazes

        if has_plane_gaze:
            planes_setup: dict[str, dict[str]] = {}
            for p in planes:
                p_def = [pl for pl in study_config.planes if pl.name==p][0]
                planes_setup[p] = plane.get_plane_from_definition(p_def, config_dir/p)

        # Read plane poses, if available
        plane_files = [working_dir/f'{naming.plane_pose_prefix}{p}.tsv' for p in planes]
        poses = {p:gt_pose.read_dict_from_file(f) for p,f in zip(planes,plane_files) if f.is_file()}
        has_plane_pose = not not poses
    else:
        raise ValueError(f'recording type "{rec_def.type}" is not understood')

    # get camera calibration info
    cam_params = ocv.CameraParams.read_from_file(working_dir / gt_naming.scene_camera_calibration_fname)

    # get previous interval coding, if available
    coding_file = working_dir / naming.coding_file
    if coding_file.is_file():
        episodes = episode.list_to_marker_dict(episode.read_list_from_file(coding_file), study_config.episodes_to_code)
    else:
        episodes = episode.get_empty_marker_dict(study_config.episodes_to_code)
    episodes_to_code = {e for e in episodes}
    # trial episodes are gotten from the reference recording if there is one and this is not the reference recording
    if study_config.sync_ref_recording and rec_def.name!=study_config.sync_ref_recording:
        # any trial coding there is should be discarded
        episodes.pop(annotation.Event.Trial, None)
        # mark trial as not codable
        if annotation.Event.Trial in episodes_to_code:
            episodes_to_code.remove(annotation.Event.Trial)
        # if there is trial coding for the reference recording, get them and show them (read only)
        if annotation.Event.Trial in study_config.episodes_to_code:
            all_recs = [r.name for r in study_config.session_def.recordings if r.name!=study_config.sync_ref_recording]
            episodes[annotation.Event.Trial] = synchronization.get_episode_frame_indices_from_ref(working_dir, annotation.Event.Trial, rec_def.name, study_config.sync_ref_recording, all_recs, study_config.sync_ref_do_time_stretch, study_config.sync_ref_average_recordings, study_config.sync_ref_stretch_which, missing_ref_coding_ok=True)
            # if nothing found, remove again so we don't have a useless empty track
            if not episodes[annotation.Event.Trial]:
                episodes.pop(annotation.Event.Trial, None)
    episodes = annotation.flatten_annotation_dict(episodes) # NB: also ensures ordering is consistent
    episodes_original = copy.deepcopy(episodes)

    # set up video playback
    # 1. timestamp info for relating audio to video frames
    video_ts = timestamps.VideoTimestamps(working_dir / gt_naming.frame_timestamps_fname)
    # 2. mediaplayer for the actual video playback, with sound if available
    ff_opts = {'volume': 1., 'sync': 'audio', 'framedrop': True}
    player = MediaPlayer(str(in_video), ff_opts=ff_opts)
    gui.set_playing(True)

    # set up annotation GUI
    gui.set_allow_pause(True)
    gui.set_allow_seek(True)
    gui.set_allow_timeline_zoom(True)
    gui.set_show_controls(True, gui.main_window_id)
    gui.set_allow_annotate(episodes_to_code, {e:_event_type_to_key_map[e] for e in episodes})
    gui.set_show_timeline(True, video_ts, episodes, gui.main_window_id)
    gui.set_show_annotation_label(True, gui.main_window_id)
    gui.set_show_action_tooltip(True, gui.main_window_id)

    # show
    sub_pixel_fac = 8   # for sub-pixel positioning
    should_exit = False
    has_requested_focus = not isMacOS # False only if on Mac OS, else True since its a no-op
    while True:
        frame, val = player.get_frame(force_refresh=True)
        if val == 'eof':
            player.toggle_pause()
        if frame is not None:
            image, pts = frame
            width, height = image.get_size()
            frame = cv2.cvtColor(np.asarray(image.to_memoryview()[0]).reshape((height,width,3)), cv2.COLOR_RGB2BGR)
            del image

        if frame is not None:
            # the audio is my shepherd and nothing shall I lack :-)
            frame_idx = video_ts.find_frame(pts*1000)  # pts is in seconds, our frame timestamps are in ms

            # if we have plane pose, draw plane origin on video
            if has_plane_pose:
                for p in planes:
                    if p in poses and frame_idx in poses[p]:
                        a = poses[p][frame_idx].get_origin_on_image(cam_params)
                        drawing.openCVCircle(frame, a, 3, (0,255,0), -1, sub_pixel_fac)
                        drawing.openCVLine(frame, (a[0],a[1]-10), (a[0],a[1]+10), (0,255,0), 1, sub_pixel_fac)
                        drawing.openCVLine(frame, (a[0]-10,a[1]), (a[0]+10,a[1]), (0,255,0), 1, sub_pixel_fac)

            # if have gaze for this frame, draw it
            # NB: usually have multiple gaze samples for a video frame, draw one
            if has_gaze:
                if frame_idx in gazes:
                    gazes[frame_idx][0].draw(frame, cam_params, sub_pixel_fac)

            # if have gaze in world info, draw it too (also only first sample)
            if has_plane_gaze:
                # first collect for which plane to draw it
                this_plane_gazes: dict[str, gaze_worldref.Gaze] = {}
                for p in planes:
                    if p in plane_gazes and frame_idx in plane_gazes[p]:
                        this_plane_gazes[p] = gaze_worldref.distance_from_plane(plane_gazes[p][frame_idx][0], planes_setup[p])
                # get best gaze (closest to a plane)
                best = None if not this_plane_gazes else sorted(this_plane_gazes.keys(), key=lambda d: this_plane_gazes[d] if this_plane_gazes[d]<=study_config.mapped_video_gaze_to_plane_margin else np.inf)
                # check if gaze is not too far outside all planes, draw
                if best is not None:
                    p = best[0]
                    plane_gazes[p][frame_idx][0].draw_on_world_video(frame, cam_params, sub_pixel_fac, None if not p in poses or not frame_idx in poses[p] else poses[p][frame_idx])

            if frame is not None:
                gui.update_image(frame, pts, frame_idx, window_id=gui.main_window_id)

        if gui.is_running() and not has_requested_focus:
            AppKit.NSApplication.sharedApplication().activateIgnoringOtherApps_(1)
            has_requested_focus = True

        requests = gui.get_requests()
        for r,p in requests:
            match r:
                case 'toggle_pause':
                    player.toggle_pause()
                    if not player.get_pause():
                        player.seek(0)  # needed to get frames rolling in again, apparently, after seeking occurred while paused
                    gui.set_playing(not player.get_pause())
                case 'seek':
                    player.seek(p, relative=False)
                case 'delta_frame':
                    new_ts = video_ts.get_timestamp(frame_idx+p)
                    if new_ts != -1.:
                        step = (new_ts-video_ts.get_timestamp(max(0,frame_idx)))/1000
                        player.seek(pts+step, relative=False)
                case 'delta_time':
                    player.seek(pts+p, relative=False)
                case 'add_coding':
                    event,frame_idx = p
                    if frame_idx not in episodes[event]:
                        episodes[event].append(frame_idx)
                        episodes[event].sort()
                        gui.notify_annotations_changed()
                case 'delete_coding':
                    event,frame_idxs = p
                    episodes[event] = [i for i in episodes[event] if i not in frame_idxs]
                    gui.notify_annotations_changed()
                case 'exit':
                    should_exit = True
                    break
        if should_exit:
            break

    player.close_player()
    gui.stop()

    # early exit if nothing has changed
    if episodes==episodes_original:
        if session.get_action_states(working_dir, True)[process.Action.CODE_EPISODES]==process_pool.State.Completed:
            return
        session.update_action_states(working_dir, process.Action.CODE_EPISODES, process_pool.State.Completed, study_config, unchanged=True)
        return

    # store coded intervals to file
    if study_config.sync_ref_recording and rec_def.name!=study_config.sync_ref_recording:
        # any (read only) trial coding there is should not be written to file
        episodes.pop(annotation.Event.Trial, None)
    episode.write_list_to_file(episode.marker_dict_to_list(episodes), coding_file)

    # update state
    session.update_action_states(working_dir, process.Action.CODE_EPISODES, process_pool.State.Completed, study_config)